#!/usr/bin/perl -I. -w
use strict;

#  ---------------------------------------------------------------------
#   configuration script for HSPICS OAI v2.0 
#    v1.1
#    June 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------

# configure the OAI data provider layer


use lib '../lib/';

use Pure::EZXML;
use Pure::X2D;


# get a single configuration option from the parsed XML file
sub getOption
{
   my ($doc, $option, $embed) = @_;
   
   if (defined $embed)
   {
      my @embedparts = split ('/', $embed);
      for ( my $i=0; $i<=$#embedparts; $i++ )
      {
         my $element = $doc->getElementsByTagName ($embedparts[$i], 0);
         if ($element->getLength > 0)
         {
            $doc = $element->item(0);
         }
         else
         {
            return '';
         }
      }
   }
   
   my $element = $doc->getElementsByTagName ($option, 0);
   if ($element->getLength > 0)
   {
      my $instance = $element->item(0);
      return $instance->getChildNodes->toString;
   }
   else
   {
      return '';
   }
}


# read in configuration from config.xml and set best-guess defaults for 
# missing values
sub ReadConfig
{
   my ($configname) = @_;
   
   my $configfile = "$configname/config.xml";
   my $config = new Pure::X2D ($configfile);
   
   # get user information from system programs
   my $hostname = `hostname`; 
   my $username = `whoami`;
   if (defined $hostname)
   {
      $hostname =~ s/[\r\n\t\s]//go;
   }
   if (defined $username)
   {
      $username =~ s/[\n\r\t\s]//go;
   }
   my $creator = `finger $username | head -1 | cut -f 4 | cut -f 2- -d ':' | cut -d ' ' -f 2-`;
   if (defined $creator)
   {
      $creator =~ s/[\n\r]//go;
   }

   if ((! defined $config->{'repositoryName'}) || ($config->{'repositoryName'}->[0] eq ''))
   {
      $config->{repositoryName} = [ "$configname Picture Archive" ];
   }

   if ((! defined $config->{'adminEmail'}) || ($config->{'adminEmail'}->[0] eq ''))
   {
      $config->{'adminEmail'} = [ '' ];

      # guess admin id from hostname and user logged in
      if ((defined $hostname) && ($hostname ne '') && 
          (defined $username) && ($username ne ''))
      {
         $hostname =~ s/[\r\n\t\s]//go;
         $username =~ s/[\n\r\t\s]//go;
         
         $config->{'adminEmail'} = [ $username.'@'.$hostname ];
      }
   }

   if ((! defined $config->{'archiveId'}) || ($config->{'archiveId'}->[0] eq ''))
   { $config->{'archiveId'} = [ $configname.'pics' ]; }

   if ((! defined $config->{'picdir'}) || ($config->{'picdir'}->[0] eq ''))
   {
      $config->{'picdir'} = [ '' ];
      if ((defined $username) && ($username ne '')) 
      {
         $config->{'picdir'} = [ "/home/$username/public_html/pictures" ]; 
      }
   }

   if ((! defined $config->{'urlprefix'}) || ($config->{'urlprefix'}->[0] eq ''))
   { 
      $config->{'urlprefix'} = [ '' ];
      if ((defined $hostname) && ($hostname ne '') &&
          (defined $username) && ($username ne ''))
      {
         $config->{'urlprefix'} = [ "http://$hostname/~$username/pictures" ];
      }
   }

   if ((! defined $config->{'publisher'}) || ($config->{'publisher'}->[0] eq ''))
   { $config->{'publisher'} = [ $configname ]; }

   if ((! defined $config->{'creator'}) || ($config->{'creator'}->[0] eq ''))
   { 
      $config->{'creator'} = [ '' ];
      if ((defined $creator) && ($creator ne ''))
      {
         $config->{'creator'} = [ $creator ]; 
      }
   }

   if ((! defined $config->{'language'}) || ($config->{'language'}->[0] eq ''))
   { $config->{'language'} = [ 'en-us' ]; }

   if ((! defined $config->{'relation'}) || ($config->{'relation'}->[0] eq ''))
   { $config->{'relation'} = [ "http://www.$configname.org" ]; }

   $config->{'mdorder'} = [ qw ( repositoryName adminEmail archiveId picdir urlprefix publisher creator language relation ) ];

   $config;
}


# write configuration data and template files
sub WriteConfig 
{
   my ($config, $configname) = @_;
   
   mkdir ($configname);
   
   open (CONFIG, ">$configname/config.xml");
   print CONFIG "<?xml version=\"1.0\" ?>\n\n";
   print CONFIG "<hspics xmlns=\"http://oai.dlib.vt.edu/~jcdlpix/metadata/OAI2config\"\n".
                "    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n".
                "    xsi:schemaLocation=\"http://oai.dlib.vt.edu/~jcdlpix/metadata/OAI2config\n".
                "                         http://oai.dlib.vt.edu/~jcdlpix/metadata/OAI2config.xsd\"\n".
                ">\n\n";
   print CONFIG $config->toXML;
   print CONFIG "</hspics>\n";
   close (CONFIG);
   
   system ("cp template/* $configname/");
}


my $fastforward = 0;

# input a line of data, with prompt and default value
sub InputLine
{
   my ($variable, $prompt) = @_;
   
   if (defined $prompt)
   {
      print "$prompt [$$variable] : ";
   }
   if ($fastforward == 1)
   {
      print "$$variable\n";
      select(undef, undef, undef, 0.25);
   }
   else
   {
      my $line = <STDIN>;
      chomp $line;
      if ($line eq '&continue')
      {
         $fastforward = 1;
         print "$$variable\n";
      }
      elsif ($line eq '&delete')
      {
         $$variable = '';
         InputLine ($variable, $prompt);
      }
      elsif ($line ne '')
      {
         $$variable = $line;
      }
   }
}


# wait for user to press ENTER
sub InputEnter
{
   my $temp = '';
   InputLine (\$temp);
}


# check validity of input and print error if necessary
sub InvalidInput
{
   my ($check, $errormsg) = @_;
   
   if ($check)
   {
      print "$errormsg\n";
      $fastforward = 0;
      return 1;
   }
   0;
}



# main program body !
sub main 
{
   $| = 1;
   my $ok;
   
   print "+-----------------------------------------+\n".
         "| Picture Engine OAI 2 Configurator       |\n".
         "+-----------------------------------------+\n".
         "| Version 1.1 :: June 2002                |\n".
         "| Hussein Suleman <hussein\@vt.edu>        |\n".
         "| Digital Library Research Laboratory     |\n".
         "| www.dlib.vt.edu :: Virginia Tech        |\n".
         "------------------------------------------+\n\n";
         
   if (! exists $ARGV[0])
   {
      print "Missing configuration name\n\n";
      print "Syntax: configure.pl <configname>\n\n";
      exit;
   }
   
   my $config = ReadConfig ($ARGV[0]);
   
   print "Defaults/previous values are in brackets - press <enter> to accept those\n".
         "enter \"&delete\" to erase a default value\n".
         "enter \"&continue\" to skip further questions and use all defaults\n".
         "press <ctrl>-c to escape at any time (new values will be lost)\n".
         "\nPress <enter> to continue\n\n";
         
   InputEnter;

         
   # get user information from system programs
   my $hostname = `hostname`; 
   my $username = `whoami`;
   if (defined $hostname)
   {
      $hostname =~ s/[\r\n\t\s]//go;
   }
   if (defined $username)
   {
      $username =~ s/[\n\r\t\s]//go;
   }
   my $creator = `finger $username | head -1 | cut -f 4 | cut -f 2- -d ':' | cut -d ' ' -f 2-`;
   if (defined $creator)
   {
      $creator =~ s/[\n\r]//go;
   }


   # Repository Name
   print "\n[REPOSITORY NAME]\n";
   print "For identification purposes, you need to specify the full name of\n".
         "the component\n".
         "Example: $ARGV[0] Picture Archive\n\n";
   do {
      InputLine (\$config->{'repositoryName'}->[0], "Enter your repository name");
   } while (InvalidInput (($config->{'repositoryName'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));
   

   # Admin Email
   print "\n[ADMINISTRATOR EMAIL]\n";
   print "When your archive is used by others, any problems\n".
         "will be reported to an administator, whose email must be provided\n".
         "Examples: admin\@oai.org, provost\@university.edu.\n\n";
   do {
      InputLine (\$config->{'adminEmail'}->[0], "Enter your administrator's email");
   } while (InvalidInput (($config->{'adminEmail'}->[0] !~ /^[^@]+@[^@\.]+\.(.+)/), 
                          "That doesnt seem to be in the right format. Please try again."));


   # Archive Id
   print "\n[ARCHIVE ID]\n";
   print "What is the archive identifier of this archive ?\n".
         "Examples: picture, $ARGV[0]\n\n";
   do {
      InputLine (\$config->{'archiveId'}->[0], "Archive ID");
   } while (InvalidInput (($config->{'archiveId'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Picture Directory
   print "\n[PICTURE DIRECTORY]\n";
   print "What is the directory that contains all the picture albums ?\n".
         "Example: /home/$username/public_html/pictures\n\n";
   do {
      InputLine (\$config->{'picdir'}->[0], "Picture Directory");
   } while (InvalidInput (($config->{'picdir'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # URL Prefix
   print "\n[URL PREFIX]\n";
   print "What is the prefix for all picture URLs ?\n".
         "Examples: http://$hostname/~$username/pictures\n\n";
   do {
      InputLine (\$config->{'urlprefix'}->[0], "URL Prefix");
   } while (InvalidInput (($config->{'urlprefix'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Publisher
   print "\n[PUBLISHER]\n";
   print "Who is the publisher of these pictures ?\n".
         "Examples: big daddy software, $ARGV[0]\n\n";
   do {
      InputLine (\$config->{'publisher'}->[0], "Publisher");
   } while (InvalidInput (($config->{'publisher'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Creator
   print "\n[CREATOR]\n";
   print "Who is the creator of these pictures ?\n".
         "Examples: big daddy, $creator\n\n";
   do {
      InputLine (\$config->{'creator'}->[0], "Creator");
   } while (InvalidInput (($config->{'creator'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Language
   print "\n[LANGUAGE]\n";
   print "What is the language of this archive's metadata ?\n".
         "Examples: en-us, es\n\n";
   do {
      InputLine (\$config->{'language'}->[0], "Language");
   } while (InvalidInput (($config->{'language'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Relation
   print "\n[RELATION]\n";
   print "What website are these pictures associated with ?\n".
         "Examples: http://www.$ARGV[0].org\n\n";
   InputLine (\$config->{'relation'}->[0], "Relation");


   WriteConfig ($config, $ARGV[0]);
   
   print "\n\n\nFinis.\n\n";
   
   print "Now you are ready to use the picture engine OAI 2 interface\n".
         "Use $ARGV[0]/oai.pl as the last bit of the baseURL for the archive interface\n\n";
}

main;
