#!/usr/bin/perl -w -I.
use strict;

#  ----------------------------------------------------------------------
# | Open Archives Initiative Harvesting Framework version 1.1            |
# | Hussein Suleman                                                      |
# | September 2001                                                       |
#  ----------------------------------------------------------------------
# |  Virginia Polytechnic Institute and State University                 |
# |  Department of Computer Science                                      |
# |  Digital Library Research Laboratory                                 |
#  ----------------------------------------------------------------------

# specify the metadata formats and the mapping from the source
# data into each of the formats

package NDLTDMetadata;

sub new
{
   my ($classname) = @_;

   my $self = {
      metadatanamespace => {
         'oai_dc'       => 'http://purl.org/dc/elements/1.1/',
         'oai_marc'     => 'http://www.openarchives.org/OAI/1.1/oai_marc',
         'oai_rfc1807'  => 'http://info.internet.isi.edu:80/in-notes/rfc/files/rfc1807.txt',
         'oai_etdms'    => 'http://www.ndltd.org/standards/metadata/etdms/1.0/'
      },
      metadataschema => {
         'oai_dc'       => 'http://www.openarchives.org/OAI/1.1/dc.xsd',
         'oai_marc'     => 'http://www.openarchives.org/OAI/1.1/oai_marc.xsd',
         'oai_rfc1807'  => 'http://www.openarchives.org/OAI/1.1/rfc1807.xsd',
         'oai_etdms'    => 'http://www.ndltd.org/standards/metadata/etdms/1.0/etdms.xsd'
      },
      metadataroot => {
         'oai_dc'       => 'dc',
         'oai_marc'     => 'oai_marc',
         'oai_rfc1807'  => 'rfc1807',
         'oai_etdms'    => 'thesis'
      },
      metadatarootparameters => {
         'oai_dc'       => '',
         'oai_marc'     => "status=\"n\" type=\"t\" level=\"m\" encLvl=\"u\" catForm=\"u\"",
         'oai_rfc1807'  => '',
         'oai_etdms'    => ''
      }
   };
   return bless $self, $classname;
}

sub MakeMIMETypes
{
   my ($self, $formats) = @_;

   # convert filenames to MIME types
   my %mimemap = (
      'txt'  => 'text/plain',
      'dvi'  => 'application/x-dvi',
      'jpg'  => 'image/jpeg',
      'doc'  => 'application/msword',
      'bmp'  => 'image/bmp',
      'aiff' => 'audio/x-aiff',
      'qt'   => 'video/quicktime',
      'mov'  => 'video/quicktime',
      'wav'  => 'audio/x-wav',
      'exe'  => 'application/octet-stream',
      'html' => 'text/html',
      'mpg'  => 'video/mpeg',
      'dtd'  => 'application/xml-dtd',
      'pdf'  => 'application/pdf',
      'dxf'  => 'image/vnd.dxf',
      'tgz'  => 'application/x-gzip',
      'htm'  => 'text/html',
      'zip'  => 'application/zip',
      'avi'  => 'video/x-msvideo',
      'ppt'  => 'application/vnd.ms-powerpoint',
      'gif'  => 'image/gif',
      'tar'  => 'application/x-tar',
      'xls'  => 'application/vnd.ms-excel',
      'tif'  => 'image/tiff',
      'xml'  => 'text/xml'
   );
   my %temptypes = ();
   foreach my $atype (@$formats)
   {
      # convert to lowercase
      $atype = lc ($atype);
      # chop off everything but the extension
      $atype =~ s/(.*)\.(.*)$/$2/g;
      if ($atype ne '')
      {
         if (defined $mimemap{$atype})
         { $temptypes{$mimemap{$atype}} = 1; }
         else
         { $temptypes{'application/octet-stream'} = 1; }
      }
   }
   return [ keys %temptypes ];
}

sub FormatDC
{
   my ($self, $hashref) = @_;

   # convert filenames to MIME type
   my $formats = $self->MakeMIMETypes ($hashref->{'format'});

   {
      'title'       => $hashref->{'title'},
      'creator'     => $hashref->{'creator'},
      'subject'     => $hashref->{'department'},
      'description' => $hashref->{'abstract'},
      'contributor' => [ map { $self->{'utility'}->lclean (${$_}[0]) } @{$hashref->{'committee'}} ],
      'publisher'   => $hashref->{'publisher'},
      'date'        => $hashref->{'date'},
      'type'        => $hashref->{'type'},
      'format'      => $formats,
      'identifier'  => $hashref->{'url'},
      'source'      => $hashref->{'url'},
      'language'    => $hashref->{'language'},
      'rights'      => [ $hashref->{'availability'}, $hashref->{'copyright_statement'} ],
      'mdorder'     => [ qw (title creator subject description contributor publisher date type format identifier source language rights) ]
   };
}

sub FormatETDMS
{
   my ($self, $hashref) = @_;

   # prepare ETDMS fields
   my $degreename = 'other';
   if (($hashref->{'dtype'} =~ /dissertation/i) || 
       ($hashref->{'dtype'} =~ /PHD/i))
   { $degreename = 'doctoral'; }
   elsif (($hashref->{'dtype'} =~ /thesis/i) || 
          ($hashref->{'dtype'} =~ /MS/i))
   { $degreename = 'masters'; }

   # convert filenames to MIME type
   my $formats = $self->MakeMIMETypes ($hashref->{'format'});

   my %oai_etdmsMap = (
      'title'       => $hashref->{'title'},
      'creator'     => $hashref->{'creator'},
      'subject'     => [ map { $self->{'utility'}->lclean ($_) } @{$hashref->{'keywords'}} ],
      'description' => $hashref->{'abstract'},
      'publisher'   => $hashref->{'publisher'},
      'contributor' => [ map { my $mapt; 
                             if (${$_}[1] =~ /chair/i) 
                             { $mapt='chair' } 
                             else
                             { $mapt='committee_member' }; 
                             [ { role=>$mapt }, $self->{'utility'}->lclean (${$_}[0]) ] 
                           } @{$hashref->{'committee'}} ],
      'date'        => $hashref->{'date'},
      'type'        => 'Electronic Thesis or Dissertation',
      'format'      => $formats,
      'identifier'  => $hashref->{'url'},
      'language'    => $hashref->{'language'},
      'rights'      => [ $hashref->{'availability'}, $hashref->{'copyright_statement'} ],
      'degree'      => {
         'name'       => $hashref->{'degree'}, 
         'level'      => $degreename,
         'discipline' => $hashref->{'department'},
         'grantor'    => $hashref->{'publisher'},
         'mdorder'    => [ qw (name level discipline grantor) ]
      },
      'mdorder' => [ qw ( title creator subject description publisher contributor date type format identifier language coverage rights degree ) ]
   );

   # enforce mandatory fields for etdms
   if ($oai_etdmsMap{'title'} eq '')
   { $oai_etdmsMap{'title'} = 'Title'; }
   if ($oai_etdmsMap{'creator'} eq '')
   { $oai_etdmsMap{'creator'} = 'Author'; }
   if ($oai_etdmsMap{'date'} eq '')
   { $oai_etdmsMap{'date'} = '2000-01-01'; }
   if ($#{$oai_etdmsMap{'subject'}} == -1)
   { $oai_etdmsMap{'subject'} = 'none'; }

   \%oai_etdmsMap;
}

sub FormatMARC
{
   my ($self, $hashref) = @_;
 
   # prepare MARC fields
    my $year = $hashref->{'date'};
    $year =~ s/^(\d{4}).+$/$1/;
    my $nfi = 0;
    my $title = $hashref->{'title'};
    if ($title =~ /^["']*(a|an|the)/i)
    { $nfi = length ($1); }
   # reduce the length of the abstract if necessary...
    my $marcabstract = $self->{'utility'}->straighten ($hashref->{'abstract'});
    if (length($marcabstract) > 9999) 
    {
       my $ellipsis = " [ abstract truncated ]...";
       $marcabstract = substr($marcabstract, 0, 9999 - length($ellipsis)) . $ellipsis;
    }
   
   {
      'fixfield' => [
         [
            { 'id' => "008" },
            "\"    s$year    vau|    sm    00| 0|eng d\""
         ]
      ],
      'varfield' => [
         [
            { 'id' => "035", 'i1' => " ", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, "($hashref->{'publisherinitials'}) $hashref->{'urn'}" ]] } 
         ],
         [
            { 'id' => "100", 'i1' => "1", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, $hashref->{'creator'} ]] } 
         ],
         [
            { 'id' => "245", 'i1' => "0", 'i2' => "$nfi" },
            { 'subfield' => [[ { 'label' => "a" }, $title ]] }
         ],
         [
            { 'id' => "256", 'i1' => " ", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, "Computer data - ".$#{$hashref->{'files'}}." files(s)" ]] }
         ],
         [
            { 'id' => "260", 'i1' => " ", 'i2' => " " },
            { 
              'subfield' => [[ { 'label' => "a" }, $hashref->{'publisherlocation'} ],
                             [ { 'label' => "b" }, "$hashref->{'publisherauthority'}, $hashref->{'publisher'}" ],
                             [ { 'label' => "c" }, $year ]]
            }
         ],
         [
            { 'id' => "502", 'i1' => " ", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, "Thesis (".$hashref->{'degree'}.")--$hashref->{'publisher'}, $hashref->{'date'}" ]] }
         ],
         [
            { 'id' => "504", 'i1' => " ", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, "Includes bibliographical references." ]] }
         ],
         [
            { 'id' => "520", 'i1' => "3", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, $marcabstract ]] }
         ],
         [
            { 'id' => "538", 'i1' => " ", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, "System requirements: PC, World Wide Web Browser and PDF reader" ]] }
         ],
         [
            { 'id' => "538", 'i1' => " ", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, "Available electronically via Internet" ]] }
         ],
         (map {
               [ 
                 { 'id' => "653", 'i1' => " ", 'i2' => " " }, 
                 { 'subfield' => [[ { 'label' => "a" }, $self->{'utility'}->lclean($_) ]] }
               ]
             } @{$hashref->{'keywords'}}),
         [
            { 'id' => "710", 'i1' => "2", 'i2' => " " },
            { 'subfield' => [[ { 'label' => "a" }, $hashref->{'publisher'} ],
                             [ { 'label' => "b" }, $hashref->{'department'} ]] }
         ],
         (map {
               [ 
                 { 'id' => "720", 'i1' => "1", 'i2' => " " }, 
                 { 
                   'subfield' => [[ { 'label' => "a" }, $self->{'utility'}->lclean(${$_}[0]) ],
                                    map { [ { 'label' => "e" }, $self->{'utility'}->lclean($_) ] } (grep { $_ && $_ !~ /^none$/i } ${$_}[1]) ] 
                 }
               ]
             } @{$hashref->{'committee'}}),
         [
            { 'id' => "856", 'i1' => "4", 'i2' => "0" },
            {
              'subfield' => [[ { 'label' => "z" }, $hashref->{'availability'} ],
                             [ { 'label' => "u" }, $hashref->{'url'} ]] 
            }
         ]
      ],
      'mdorder' => [ qw (fixfield varfield) ]
   };
}

sub FormatRFC1807
{
   my ($self, $hashref) = @_;

   {
      'bib-version'   => '1',
      'id'            => $hashref->{'urn'},
      'entry'         => $hashref->{'date'},
      'organization'  => $hashref->{'publisher'},
      'title'         => $hashref->{'title'},
      'type'          => "Thesis/Dissertation",
      'author'        => $hashref->{'creator'},
      'other_access'  => "url:".$hashref->{'url'},
      'language'      => $hashref->{'language'},
      'abstract'      => $hashref->{'abstract'},
      'mdorder'       => [ qw (bib-version id entry organization title type author other_access language abstract) ]
   };
}

sub MetadataFormatisValid
{
   my ($self, $metadataFormat) = @_;
   my $i;
   my $found = 0;
   foreach $i (keys %{$self->{'metadatanamespace'}})
   {
      if ($metadataFormat eq $i)
      { $found = 1; }
   }
   if ($found == 1)
   { return 1; }
   else
   { return 0; }
}

sub Format
{
   my ($self, $hashref, $metadataFormat) = @_;
   
   # cater for missing fields in earlier versions of ETD-db
   ## add in copyright_statement for versions prior to 1.6a 
   if (! defined $hashref->{'copyright_statement'})
   {
      $hashref->{'copyright_statement'} = 'Copyright information available at source archive';
   }
   
   # set hard defaults for fields
   foreach my $key (qw (urn title last_name first_name middle_name suffix department abstract adate files type format url availability degree dtype))
   {
      if (! defined $hashref->{$key})
      {
         $hashref->{$key} = '';
      }
   }
   if (! defined $hashref->{'committee'}) { $hashref->{'committee'} = ['', ''] };

   # perform a little bit of pre-cleaning of fields
   $hashref->{'title'}      = $self->{'utility'}->lclean ($hashref->{'title'});
   $hashref->{'creator'}    = $self->{'utility'}->lclean ($hashref->{'last_name'}.", ".join (' ', grep {$_ ne ''} ($hashref->{'first_name'}, $hashref->{'middle_name'}, $hashref->{'suffix'})));
   $hashref->{'department'} = $self->{'utility'}->lclean ($hashref->{'department'});
   $hashref->{'abstract'}   = $self->{'utility'}->pclean ($hashref->{'abstract'});
   $hashref->{'date'}       = $hashref->{'adate'};
   $hashref->{'type'}       = "text";
   $hashref->{'format'}     = $hashref->{'files'};
   $hashref->{'url'}        = $self->{'utility'}->lclean ($hashref->{'url'});
   $hashref->{'availability'} = $self->{'utility'}->lclean ($hashref->{'availability'}); 
   $hashref->{'copyright_statement'} = $self->{'utility'}->lclean ($hashref->{'copyright_statement'});
   $hashref->{'degree'}     = $self->{'utility'}->lclean ($hashref->{'degree'});

   my $output = 
      "<record>\n".
      "<header>\n".
      "<identifier>oai:$hashref->{'archiveId'}:$hashref->{'urn'}</identifier>\n".
      "<datestamp>$hashref->{'date'}</datestamp>\n".
      "</header>\n";

   if ($self->MetadataFormatisValid ($metadataFormat))
   {
      $output .=
         "<metadata>\n".
         "<$self->{'metadataroot'}->{$metadataFormat} ".
         "xmlns=\"$self->{'metadatanamespace'}->{$metadataFormat}\" ".
         "xsi:schemaLocation=\"$self->{'metadatanamespace'}->{$metadataFormat} ".
         "$self->{'metadataschema'}->{$metadataFormat}\" ".
         "$self->{'metadatarootparameters'}->{$metadataFormat}>\n";

      my $outputmap;
      if ($metadataFormat eq 'oai_dc')
      { $outputmap = $self->FormatDC ($hashref); }
      elsif ($metadataFormat eq 'oai_marc')
      { $outputmap = $self->FormatMARC ($hashref); }
      elsif ($metadataFormat eq 'oai_rfc1807')
      { $outputmap = $self->FormatRFC1807 ($hashref); }
      elsif ($metadataFormat eq 'oai_etdms')
      { $outputmap = $self->FormatETDMS ($hashref); }
      $output .= $self->{'utility'}->FormatXML ($outputmap);

      $output .=
         "</$self->{'metadataroot'}->{$metadataFormat}>\n".
         "</metadata>\n";
   }

   $output."</record>\n";
}

1;
