#!/usr/bin/perl -I. -w
use strict;

#  ---------------------------------------------------------------------
#   configuration script for DBUnion v1.1 
#    v1.1
#    August 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------

# configure the OAI data provider layer


use lib '../lib/';

use Pure::EZXML;
use Pure::EZHTTP;
use Pure::X2D;


# get data through HTTP, with retries
sub getHTTP
{
   my ($url) = @_;

   # create a user agent object
   my $ua = new Pure::EZHTTP;
   $ua->agent("VT OAI Harvester/1.0 " . $ua->agent);

   # create a request
   my $req = new Pure::EZHTTP GET => $url;

   my $state = 0;
   my $res;
   while ($state == 0)
   {
      # pass request to the user agent and get a response back
      $res = $ua->request($req);
      
      if ($res->code == 503)
      {
         my $sleep = $res->header ('Retry-After');
         if (not defined ($sleep) || ($sleep < 0) || ($sleep > 86400))
         { $state = 1;}
         else
         { sleep ($sleep); }
      }
      else
      { $state = 1; }
   }
   
   # return results to caller
   $res;
}


# get a single configuration option from the parsed XML file
sub getOption
{
   my ($doc, $option, $embed) = @_;
   
   if (defined $embed)
   {
      my @embedparts = split ('/', $embed);
      for ( my $i=0; $i<=$#embedparts; $i++ )
      {
         my $element = $doc->getElementsByTagName ($embedparts[$i], 0);
         if ($element->getLength > 0)
         {
            $doc = $element->item(0);
         }
         else
         {
            return '';
         }
      }
   }
   
   my $element = $doc->getElementsByTagName ($option, 0);
   if ($element->getLength > 0)
   {
      my $instance = $element->item(0);
      return $instance->getChildNodes->toString;
   }
   else
   {
      return '';
   }
}


# read in configuration from cfg.pl and set best-guess defaults for 
# missing values
sub ReadConfig
{
   my ($configname) = @_;
   
   my $configfile = "$configname/config.xml";
   my $config = new Pure::X2D ($configfile);
   
   if ((! defined $config->{'repositoryName'}) || ($config->{'repositoryName'}->[0] eq ''))
   {
      $config->{repositoryName} = [ 'ODL Union Engine' ];
   }

   if ((! defined $config->{'adminEmail'}) || ($config->{'adminEmail'}->[0] eq ''))
   {
      $config->{'adminEmail'} = [ '' ];

      # guess admin id from hostname and user logged in
      my $hostname = `hostname`; 
      my $username = `whoami`;
      if ((defined $hostname) && ($hostname ne '') && 
          (defined $username) && ($username ne ''))
      {
         $hostname =~ s/[\r\n\t\s]//go;
         $username =~ s/[\n\r\t\s]//go;
         
         $config->{'adminEmail'} = [ $username.'@'.$hostname ];
      }
   }

   if ((! defined $config->{'database'}) || ($config->{'database'}->[0] eq ''))
   { $config->{'database'} = [ 'odlunion' ]; }
   if ((! defined $config->{'dbusername'}) || ($config->{'dbusername'}->[0] eq ''))
   {
      $config->{'dbusername'} = [ '' ];
      
      my $username = `whoami`;
      if ((defined $username) && ($username ne ''))
      {
         $username =~ s/[\r\t\n\s]//go;
         $config->{'dbusername'} = [ $username ]; 
      }
   }
   if ((! defined $config->{'dbpassword'}) || ($config->{'dbpassword'}->[0] eq ''))
   { $config->{'dbpassword'} = [ '' ]; }
   if ((! defined $config->{'table'}) || ($config->{'table'}->[0] eq ''))
   { $config->{'table'} = [ $configname ]; }
   if ((! defined $config->{'recordlimit'}) || ($config->{'recordlimit'}->[0] eq ''))
   { $config->{'recordlimit'} = [ '' ]; }
   
   $config->{'mdorder'} = [ qw ( repositoryName adminEmail database dbusername dbpassword table recordlimit archive ) ];

   if (! defined $config->{'archive'})
   { $config->{'archive'} = []; }

   foreach my $archive (@{$config->{'archive'}})
   {
      if ((! defined $archive->{'identifier'}) || ($archive->{'identifier'} eq ''))
      { $archive->{'identifier'} = [ $configname ]; }
      if ((! defined $archive->{'url'}) || ($archive->{'url'} eq ''))
      { $archive->{'url'} = [ '' ]; }
      if ((! defined $archive->{'metadataPrefix'}) || ($archive->{'metadataPrefix'} eq ''))
      { $archive->{'metadataPrefix'} = [ 'oai_dc' ]; }
      if ((! defined $archive->{'interval'}) || ($archive->{'interval'} eq ''))
      { $archive->{'interval'} = [ 86400 ]; }
      if ((! defined $archive->{'interrequestgap'}) || ($archive->{'interrequestgap'} eq ''))
      { $archive->{'interrequestgap'} = [ 10 ]; }
      if ((! defined $archive->{'set'}) || ($archive->{'set'} eq ''))
      { $archive->{'set'} = [ '' ]; }
      if ((! defined $archive->{'overlap'}) || ($archive->{'overlap'} eq ''))
      { $archive->{'overlap'} = [ 0 ]; }
      if ((! defined $archive->{'granularity'}) || ($archive->{'granularity'} eq ''))
      { $archive->{'granularity'} = [ '' ]; }

      $archive->{'mdorder'} = [ qw ( identifier url metadataPrefix interval interrequestgap set overlap granularity ) ];
   }

   $config;
}


# write configuration data and template files
sub WriteConfig 
{
   my ($config, $configname) = @_;
   
   mkdir ($configname);
   
   open (CONFIG, ">$configname/config.xml");
   print CONFIG "<?xml version=\"1.0\" ?>\n\n";
   print CONFIG "<odlunion xmlns=\"http://oai.dlib.vt.edu/ODL/DBUnion/config\"\n".
                "    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n".
                "    xsi:schemaLocation=\"http://oai.dlib.vt.edu/ODL/DBUnion/config\n".
                "                         http://oai.dlib.vt.edu/ODL/DBUnion/config.xsd\"\n".
                ">\n\n";
   print CONFIG $config->toXML;
   print CONFIG "</odlunion>\n";
   close (CONFIG);
   
   system ("cp template/* $configname/");
   mkdir ("$configname/data");
}


my $fastforward = 0;

# input a line of data, with prompt and default value
sub InputLine
{
   my ($variable, $prompt) = @_;
   
   if (defined $prompt)
   {
      print "$prompt [$$variable] : ";
   }
   if ($fastforward == 1)
   {
      print "$$variable\n";
      select(undef, undef, undef, 0.25);
   }
   else
   {
      my $line = <STDIN>;
      chomp $line;
      if ($line eq '&continue')
      {
         $fastforward = 1;
         print "$$variable\n";
      }
      elsif ($line eq '&delete')
      {
         $$variable = '';
         InputLine ($variable, $prompt);
      }
      elsif ($line ne '')
      {
         $$variable = $line;
      }
   }
}


# wait for user to press ENTER
sub InputEnter
{
   my $temp = '';
   InputLine (\$temp);
}


# check validity of input and print error if necessary
sub InvalidInput
{
   my ($check, $errormsg) = @_;
   
   if ($check)
   {
      print "$errormsg\n";
      $fastforward = 0;
      return 1;
   }
   0;
}


# check for database operations on a single table
sub CheckTable
{
   my ($dbh, $table) = @_;

   my $sth = $dbh->prepare ("select count(*) from $table");
   if (! defined $sth)
   {
      return "Error preparing a select for table $table";
   }
   if (!($sth->execute))
   {
      $sth->finish;
      return "Error executing a select on table $table";
   }
   if (!($sth->fetchrow_hashref))
   {
      $sth->finish;
      return "Error getting row from table $table";
   }
   $sth->finish;
   return '';
}


# archive configuration for a single archive
sub archive_config
{
   my ($archive) = @_;

   # Archive Identifier
   print "\n[ARCHIVE IDENTIFIER]\n";
   print "You need a unique name by which to refer to the archive you\n".
         "will harvest metadata from\n";
   print "Examples: $ARGV[0], VTETD\n\n";
   do {
      InputLine (\$archive->{'identifier'}->[0], "Archive identifier");
   } while (InvalidInput (($archive->{'identifier'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Archive URL
   my $ok = 0;
   my $gotodl = 0;
   my $gotoai2 = 0;
   my $oai2granularity = 'YYYY-MM-DD';
   print "\n[ARCHIVE URL]\n";
   print "This is the baseURL of the archive being harvested from.\n\n";
   do {
      do {
         InputLine (\$archive->{'url'}->[0], "Archive baseURL");
      } while (InvalidInput (($archive->{'url'}->[0] eq ''),
                             "This field cannot be left blank. Please enter a value."));

      print "Issuing Identify request to archive ...\n";
      
      # submit Identify to archive
      my $res = getHTTP ($archive->{'url'}->[0].'?verb=Identify');
      print "Got response code : ".$res->code."\n";
      
      if ($res->code == 200)
      {
         $ok = 1;
         
         my $parser = new Pure::EZXML;
         my $document = $parser->parse ($res->content)->getDocumentElement;
         
         foreach my $description ($document->getElementsByTagName ('description', 0))
         {
            foreach my $odldescription ($description->getElementsByTagName ('odl-description', 0))
            {
               $gotodl = 1;
            }
         }
         
         if ($document->getTagName eq 'OAI-PMH')
         {
            $gotoai2 = 1;
            
            foreach my $Identify ($document->getElementsByTagName ('Identify', 0))
            {
               foreach my $granularity ($Identify->getElementsByTagName ('granularity', 0))
               {
                  $oai2granularity = $granularity->getChildNodes->toString;
               }
            }
         }
         
         if ($archive->{'overlap'}->[0] == 0)
         {
            if ($gotodl == 1)
            {
               $archive->{'overlap'}->[0] = 86401;
            }
            elsif ($gotoai2 == 1)
            {
               if ($oai2granularity eq 'YYYY-MM-DD')
               {
                  $archive->{'overlap'}->[0] = 86400;
               }
               else
               {
                  $archive->{'overlap'}->[0] = 1;
               }
            }
            else
            {
               $archive->{'overlap'}->[0] = 172800;
            }
         }
         
         if ($archive->{'granularity'}->[0] eq '')
         {
            if ($gotodl == 1)
            {
               $archive->{'granularity'}->[0] = 'second';
            }
            elsif ($gotoai2 == 1)
            {
               if ($oai2granularity eq 'YYYY-MM-DD')
               {
                  $archive->{'granularity'}->[0] = 'day';
               }
               else
               {
                  $archive->{'granularity'}->[0] = 'second';
               }
            }
            else
            {
               $archive->{'granularity'}->[0] = 'day';
            }
         }
      }
      else
      {
         print "Error with archive or URL - please check\n";
      }
   } while ($ok == 0);


   # harvesting interval
   print "\n[HARVESTING INTERVAL]\n";
   print "Choose how often you want to harvest (in seconds)\n".
         "86400 = daily, 3600 = hourly\n\n";
   do {
      InputLine (\$archive->{interval}->[0], "Harvesting interval");
   } while (InvalidInput (($archive->{interval}->[0] < 1),
                          "This field must be a number and >0. Please try again."));


   # seconds overlap
   print "\n[HARVESTING OVERLAP]\n";
   print "You need to select the number of seconds by which to overlap\n".
         "harvesting to handle differences in times\n";
   if ($gotodl == 1)
   {
      print "It seems you are harvesting from an ODL archive !\n".
            "Use 1 for an archive in the same timezone or\n".
            "86401 for an archive in a different or unknown timezone\n\n";
   }
   elsif ($gotoai2 == 1)
   {
      print "It seems you are harvesting from an OAI-PMH v2.0 archive !\n";
      if ($oai2granularity eq 'YYYY-MM-DD')
      {
         print "Use 86400 for a single day\n\n";
      }
      else
      {
         print "Use 1 for a single second\n\n";
      }
   }
   else
   {
      print "It seems you are harvesting from a OAI-PMH v1.0/1.1 archive !\n".
            "Use 86400 for an archive in the same timezone or\n".
            "172800 for an archive in a different or unknown timezone\n\n";
   }
   do {
      InputLine (\$archive->{'overlap'}->[0], "Harvesting overlap");
   } while (InvalidInput (($archive->{'overlap'}->[0] < 1),
                          "This field must be a number and >0. Please try again."));


   # harvesting granularity
   print "\n[HARVESTING GRANULARITY]\n";
   print "This specifies whether harvesting is performed by day or second\n\n";
   if ($gotodl == 1)
   {
      print "It seems you are harvesting from an ODL archive !\n".
            "Use \'day\' or \'second\'\n\n";
   }
   elsif ($gotoai2 == 1)
   {
      if ($oai2granularity eq 'YYYY-MM-DD')
      {
         print "It seems you are harvesting from an OAI-PMH v2.0 archive with only days!\n".
               "Use \'day\'\n\n";
      }
      else
      {
         print "It seems you are harvesting from an OAI-PMH v2.0 archive with seconds!\n".
               "Use \'day\' or \'second\'\n\n";
      }
   }
   else
   {
      print "It seems you are harvesting from an OAI-PMH v1.0/1.1 archive !\n".
            "Use \'day\'\n\n";
   }
   do {
      InputLine (\$archive->{'granularity'}->[0], "Harvesting granularity");
   } while (InvalidInput ((($archive->{'granularity'}->[0] ne 'day') && ($archive->{'granularity'}->[0] ne 'second')),
                          "This field must be either day or second. Please try again."));


   # metadataPrefix
   print "\n[metadataPrefix]\n";
   print "This is the metadata format that will be harvested and stored.\n\n";
   
   # submit ListMetadataFormats to archive and generate list of formats
   my $res = getHTTP ($archive->{'url'}->[0].'?verb=ListMetadataFormats');
   if ($res->code != 200)
   {
      print "Fatal error: Archive responds to Identify but not ListMetadataFormats!\n";
      exit;
   }
   my $parser = new Pure::EZXML;
   my $lmfdoc = $parser->parse ($res->content)->getDocumentElement;
   my @mdprefixes = ();
   if ($lmfdoc->getTagName eq 'OAI-PMH')
   {
      $lmfdoc = $lmfdoc->getElementsByTagName ('ListMetadataFormats', 0)->item(0);
   }
   foreach my $mdf ($lmfdoc->getElementsByTagName ('metadataFormat', 0))
   {
      foreach my $mdp ($mdf->getElementsByTagName ('metadataPrefix', 0))
      {
         push (@mdprefixes, $mdp->getChildNodes->toString);
      }
   }
   print "Valid formats: ".join (',', @mdprefixes);
   print "\n\n";
   do {
      InputLine (\$archive->{'metadataPrefix'}->[0], "metadataPrefix");
   } while (InvalidInput ((1 - grep {$_ eq $archive->{'metadataPrefix'}->[0]} @mdprefixes),
                          "Please choose a value from the list above."));


   # set
   print "\n[set]\n";
   print "Choose a set to harvested (descriptions are in brackets).\n".
         "Leave blank to harvest the whole archive.\n\n";
   
   # submit ListSets to archive and generate list of formats
   $res = getHTTP ($archive->{'url'}->[0].'?verb=ListSets');
   if ($res->code != 200)
   {
      print "Fatal error: Archive responds to Identify but not ListSets!\n";
      exit;
   }
   my $lsdoc = $parser->parse ($res->content)->getDocumentElement;
   my $nosets = 0;
   if ($lsdoc->getTagName eq 'OAI-PMH')
   {
      if ($lsdoc->getElementsByTagName ('ListSets', 0)->getLength == 0)
      {
         $nosets = 1;
      }
      else
      {
         $lsdoc = $lsdoc->getElementsByTagName ('ListSets', 0)->item(0);
      }
   }
   my @sets = ();
   print "Valid sets:\n";
   if ($nosets == 0)
   {
      foreach my $set ($lsdoc->getElementsByTagName ('set', 0))
      {
         my $setSpec = $set->getElementsByTagName ('setSpec', 0)->item(0)->getChildNodes->toString;
         my $setName = $set->getElementsByTagName ('setName', 0)->item(0)->getChildNodes->toString;
         push (@sets, $setSpec);
         print "  $setSpec [$setName]\n";
      }
   }
   print "\n";
   do {
      InputLine (\$archive->{'set'}->[0], "set");
   } while (InvalidInput ((($archive->{'set'}->[0] ne '' ) && 
                           (1 - grep {$_ eq $archive->{'set'}->[0]} @sets)),
                          "Please choose a value from the list above."));
}


# main program body !
sub main 
{
   $| = 1;
   my $ok;
   
   print "+-----------------------------------------+\n".
         "| DBUnion Union Engine Configurator       |\n".
         "+-----------------------------------------+\n".
         "| Version 1.1 :: August 2002              |\n".
         "| Hussein Suleman <hussein\@vt.edu>        |\n".
         "| Digital Library Research Laboratory     |\n".
         "| www.dlib.vt.edu :: Virginia Tech        |\n".
         "------------------------------------------+\n\n";
         
   if (! exists $ARGV[0])
   {
      print "Missing configuration name\n\n";
      print "Syntax: configure.pl <configname>\n\n";
      exit;
   }
   
   my $config = ReadConfig ($ARGV[0]);
   
   print "Defaults/previous values are in brackets - press <enter> to accept those\n".
         "enter \"&delete\" to erase a default value\n".
         "enter \"&continue\" to skip further questions and use all defaults\n".
         "press <ctrl>-c to escape at any time (new values will be lost)\n".
         "\nPress <enter> to continue\n\n";
         
   InputEnter;

         
   # test for not finding DBI
   my $gotdbi = eval "require DBI";
   print "\n[DBI]\n";
   print "You need the DBI Perl module installed in order to use this script\n".
         "Checking ... ";
   if (! defined $gotdbi)
   {
      print "not installed !\n".
            "Please press CTRL-C and then install that module.\n".
            "You also need a module specific to your database e.g., DBD::mysql.\n".
            " \"man CPAN\" for help on how to do this.\n\n";
      exit;
   }
   else
   {
      print "ok\n\n";
   }
   

   # test for available driver
   my @dbi_drivers = DBI->available_drivers;
   my $driver_list = join (', ', @dbi_drivers);
   print "\n[DATABASE DRIVER]\n";
   print "DBI requires special drivers to access different databases. Please\n".
         "confirm that the list below contains the name of your database.\n".
         "If it is not here or you are not sure, press CTRL-C and install it\n";
   print "\n".$driver_list."\n\n";
   print "Press ENTER to continue\n";
   InputEnter;

   
   # Database Name + Username + Password
   $ok = 0;
   print "\n[DATABASE CONNECTION]\n";
   print "The ODLUnion scripts need to connect to a database to store\n".
         "and retrieve data. For this you need to supply the name of the\n".
         "database in the format required by your database\n".
         "(you will be asked for username and password next).\n".
         "If you dont have permission to access the database from this host\n".
         "you may want to install this OAI script on another machine or grant\n".
         "permissions before you continue. If the database does not exist, you\n".
         "probably should create it before continuing.\n";
   do {
      my @dbparts = ('DBI', 'mysql', 'odlunion');
      if ($config->{'database'}->[0] =~ /([^:]+):([^:]+):(.*)/)
      {
         @dbparts = ($1, $2, $3);
      }
   
      my $found = 0;
      do {
         print "\nDrivers: $driver_list\n";
         InputLine (\$dbparts[1], "Choose database driver from above list");
         foreach my $testdbidriver (@dbi_drivers)
         {
            if ($testdbidriver eq $dbparts[1])
            { $found = 1; }
         }
      } while (InvalidInput (($found == 0), "Invalid choice. Please choose again."));
      
      InputLine (\$dbparts[2], "Enter name of database");

      $dbparts[0] = 'DBI';
      $config->{'database'}->[0] = join (':', @dbparts);

      InputLine (\$config->{'dbusername'}->[0], "Enter the username to be used to connect");
      InputLine (\$config->{'dbpassword'}->[0], "Enter the password to be used to connect");

      print "Checking ... ";
      my $dbh = DBI->connect ($config->{'database'}->[0], $config->{'dbusername'}->[0], 
                              $config->{'dbpassword'}->[0], { PrintError=>0 });
      if (! defined $dbh)
      { 
         print "Cannot connect to the specified database with user/pass. Try again.\n";
         $fastforward = 0;
      }
      else
      {
         print "ok\n";
         $dbh->disconnect;
         $ok = 1;
      }
   } while ($ok == 0);
   

   # Database Table
   print "\n[DATABASE TABLE]\n";
   print "Within the database, multiple tables are needed for the indices\n".
         "Enter a unique table prefix for this component\n";
   print "Examples: dbunion, $ARGV[0]\n\n";
   do {
      InputLine (\$config->{'table'}->[0], "Table prefix");
   } while (InvalidInput (($config->{'table'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));


   # Repository Name
   print "\n[REPOSITORY NAME]\n";
   print "For identification purposes, you need to specify the full name of\n".
         "the component\n".
         "Example: JCDL ODL Union Engine\n\n";
   do {
      InputLine (\$config->{'repositoryName'}->[0], "Enter your repository name");
   } while (InvalidInput (($config->{'repositoryName'}->[0] eq ''),
                          "This field cannot be left blank. Please enter a value."));
   

   # Admin Email
   print "\n[ADMINISTRATOR EMAIL]\n";
   print "When your union archive is used by others, any problems\n".
         "will be reported to an administator, whose email must be provided\n".
         "Examples: admin\@odlunion.org, provost\@university.edu.\n\n";
   do {
      InputLine (\$config->{'adminEmail'}->[0], "Enter your administrator's email");
   } while (InvalidInput (($config->{'adminEmail'}->[0] !~ /^[^@]+@[^@\.]+\.(.+)/), 
                          "That doesnt seem to be in the right format. Please try again."));


   # Record Limit
   print "\n[RECORD LIMIT]\n";
   print "Indicate the number of records to be issued in each batch\n".
         "before a resumptionToken is issued\n\n";
   do {
      InputLine (\$config->{'recordlimit'}->[0], "Record limit");
   } while (InvalidInput ((($config->{'recordlimit'}->[0] ne '') && ($config->{'recordlimit'}->[0] < 1)),
                          "This field must be a number and >0. Please try again."));


   # Archive administration
   my $choice;
   print "\n[ARCHIVES]\n";
   print "Add all the archives that are part of this union collection\n";
   do {
      # print out archive listing
      print "\nCurrent list of archives:\n";
      if ($#{$config->{'archive'}} == -1)
      {
         print "No archives currently defined !\n";
      }
      for ( my $i=0; $i<=$#{$config->{'archive'}}; $i++ )
      {
         my $archive = $config->{'archive'}->[$i];
         print "$i. ".$archive->{'identifier'}->[0].' '.$archive->{'url'}->[0].
               ' '.$archive->{'metadataPrefix'}->[0].' '.$archive->{'set'}->[0].
               "\n";
      }
      
      # print out menu
      my $instrline = "\nSelect from: [A]dd   [R]emove   [E]dit   [D]one\n";
      if ($#{$config->{'archive'}} == -1)
      {
         $instrline = "\nSelect from: [A]dd   [D]one\n";
      }
      print $instrline;

      # get choice
      do {
         $choice = 'D';
         InputLine (\$choice, "Enter your choice");
         $choice = uc ($choice);
      } while (InvalidInput (((($choice !~ /^[ARED]$/) || ($#{$config->{'archive'}} == -1)) &&
                              (($choice !~ /^[AD]$/) || ($#{$config->{'archive'}} > -1))),
                             $instrline));

      # act on choice
      if ($choice eq 'A')
      {
         push (@{$config->{'archive'}}, { identifier => [''], url => [''], 
           metadataPrefix => [''], interval => ['86400'], interrequestgap => ['10'], 
           set => [''], overlap => ['0'], granularity => [''],
           mdorder => [ qw ( identifier url metadataPrefix interval interrequestgap set overlap granularity ) ] } );
         archive_config ($config->{'archive'}->[-1]);
      }
      elsif (($choice eq 'E') || ($choice eq 'R'))
      {
         my $archiveno = '';
         do {
            InputLine (\$archiveno, "Enter the number of the archive");
         } while (InvalidInput ((($archiveno !~ /^[0-9]+$/) || ($archiveno > $#{$config->{'archive'}})),
                                "This has to be an integer in the range listed. Please try again."));
         
         if ($choice eq 'R')
         {
            my $confirm = 'Y/N';
            do {
               InputLine (\$confirm, "Confirm");
               $confirm = uc ($confirm);
            } while (InvalidInput (($confirm !~ /^[NY]$/),
                                   "Enter Y or N"));
            if ($confirm eq 'Y')
            {
               splice (@{$config->{'archive'}}, $archiveno, 1);
            }
         }
         elsif ($choice eq 'E')
         {
            archive_config ($config->{'archive'}->[$archiveno]);
         }
      }
      
   } while ($choice ne 'D');
   
   WriteConfig ($config, $ARGV[0]);
   
   print "\n\n\nFinis.\n\n";
   
   print "Now you are ready to harvest and test the union engine\n".
         "Run $ARGV[0]/harvest.pl to try harvesting\n".
         "Use $ARGV[0]/union.pl as the last bit of the baseURL for the archive interface\n\n";
}

main;
