# +---------------------------------------------------------------------+
#    HTTP downloader in pure-perl to mimic LWP functionality
#    v1.1
#    May 2003
# +----------------------------------+----------------------------------+
# |  Hussein Suleman                 |  Department of Computer Science  |
# |   <hussein@cs.uct.ac.za>         |  University of Cape Town         |
# |   http://www.husseinsspace.com   |  Cape Town, South Africa         |
# +----------------------------------+----------------------------------+


package Pure::EZHTTP;


use IO::Socket;


# constructor
sub new
{
   my ($classname, %parmlist) = @_;
   my $self = { 
      agent     => '',
      code      => 500,
      content   => '',
      headers   => {},
      message   => '',
      proxyname => '',
      proxyport => 80,
      noproxy   => [],
   };
   if (%parmlist)
   {
      $self->{parmlist} = \%parmlist;
   }
   bless $self, $classname;
   return $self;
}


# set the name of the user agent
sub agent
{
   my ($self, $newagent) = @_;
   if (defined $newagent)
   {
      $self->{agent} = $newagent;
   }
   return $self->{agent};
}


# submit an HTTP request 
sub request
{
   my ($self, $request) = @_;
   
   # handle GET request
   if ($request->{parmlist}->{'GET'} ne '')
   {
      my $url = $request->{parmlist}->{'GET'};
      my ($host, $port, $path);

      if ($url =~ /http:\/\/([^:\/]+)(:([0-9]+))?([^#]*)(.*)/)
      {
         $host = $1;
         $port = defined $3 && $3 ne '' ? $3 : '80';
         $path = defined $4 && $4 ne '' ? $4 : '/';
      }
      
      my $EOL = "\015\012";
      my $BLANK = $EOL x 2;

      # make connection
      my $remote;
      if (($self->{'proxyname'} eq '') || ($self->must_not_proxy ($host)))
      {
         $remote = new IO::Socket::INET ( Proto=>'tcp', PeerAddr=>$host, PeerPort=>$port);
         unless ($remote) # cannot connect
         {
            return $request;
         }
      
         $remote->autoflush(1);

         print $remote "GET $path HTTP/1.0". $EOL.
                       "Host: $host" . $BLANK;
      }
      else
      {
         $remote = new IO::Socket::INET ( Proto=>'tcp', PeerAddr=>$self->{'proxyname'}, PeerPort=>$self->{'proxyport'});
         unless ($remote) # cannot connect
         {
            return $request;
         }
      
         $remote->autoflush(1);

         print $remote "GET $url HTTP/1.0". $BLANK;
      }
      
      # get return status code
      my $aline = <$remote>;
      if ($aline =~ /HTTP\/[^ ]* ([0-9]+)(.*)/)
      {
         $request->{code} = $1;
         $request->{message} = $2;
         $request->{message} =~ s/\s*(.*)/$1/;
      }
      
      # get headers
      while ($aline = <$remote>) 
      {
         if ($aline =~ /^[\s]*$/)
         {
            last;
         }
         elsif ($aline =~ /([^:]+): (.*)/)
         {
            $request->{headers}->{$1} = $2;
         }
      }
      
      # get body
      while ($aline = <$remote>) 
      {
         $request->{content} .= $aline;
      }
      
      close $remote;
   }
   
   return $request;
}


# retrieve HTTP status code
sub code
{
   my ($self) = @_;
  
   return $self->{code};
}


# retrieve HTTP status message
sub message
{
   my ($self) = @_;
  
   return $self->{message};
}


# retrieve HTTP header by name
sub header
{
   my ($self, $header) = @_;
   
   return $self->{headers}->{$header};
}


# retrieve HTTP body
sub content
{
   my ($self) = @_;
   
   return $self->{content};
}


# set a proxy server
sub proxy
{
   my ($self, $protocol, $url) = @_;

   if ($url =~ /http:\/\/([^:\/]+)(:([0-9]+))?([^#]*)(.*)/)
   {
      $self->{'proxyname'} = $1;
      $self->{'proxyport'} = defined $3 && $3 ne '' ? $3 : '80';
   }
}


# add a proxy exclusion url or nix the list
sub no_proxy
{
   my ($self, @domains) = @_;
   
   if ($#domains > -1)
   {
      push (@{$self->{'noproxy'}}, @domains);
   }
   else
   {
      $self->{'noproxy'} = [];
   }
}


# set proxy variables from environment
sub env_proxy
{
   my ($self) = @_;
   
   if (exists $ENV{'HTTP_PROXY'})
   {
      $self->proxy ('http', $ENV{'HTTP_PROXY'});
   }
   if (exists $ENV{'NO_PROXY'})
   {
      my @domains = split (',', $ENV{'NO_PROXY'});
      $self->no_proxy (@domains);
   }
}


# check if a url should be proxied
sub must_not_proxy
{
   my ($self, $domain) = @_;
   
   my $domainlength = length ($domain);
   foreach my $d (@{$self->{'noproxy'}})
   {
      my $dlength = length ($d);
      if (($dlength <= $domainlength) &&
          ($dlength > 0) &&
          ($domainlength > 0) &&
          (substr ($domain, - $dlength) eq $d))
      { 
         return 1;
      }
   }

   return 0;
}


1;


# HISTORY
# v1.1 - added proxy server support
# v1.0 - first release version

