#  ---------------------------------------------------------------------
#   General database implementation of ODL-Search (v1.0) - servive prov.  
#    v1.1
#    January 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------


package ODL::IRDB::SearchSP;


use Pure::Log;
use Pure::X2D;
use Pure::EZXML;
use ODL::IRDB::XMLIR;

use XOAI::Harvester;
use vars ('@ISA');
@ISA = ("XOAI::Harvester");


# constructor
sub new
{
   my ($classname, $configfile) = @_;

   my $self = $classname->SUPER::new ();

   # get configuration from file
   my $con = new Pure::X2D ($configfile);
   $self->{'database'} = $con->param ('database', '');
   $self->{'dbusername'} = $con->param ('dbusername', '');
   $self->{'dbpassword'} = $con->param ('dbpassword', '');
   $self->{'table'} = $con->param ('table', 'irdb');
   $self->{'xmlroot'} = 'metadata';
   $self->{'log'} = new Pure::Log ('harvest.log');
   $self->{'counter'} = 0;
   $self->{'archive'} = $con->{'archive'};

   # create IR object
   $self->{'ir'} = new ODL::IRDB::XMLIR 
     ($self->{'database'}, $self->{'dbusername'}, 
      $self->{'dbpassword'}, $self->{'table'},
      $self->{'xmlroot'});
   
   bless $self, $classname;
   return $self;
}


# destructor
sub dispose
{
   my ($self) = @_;
   $self->{'ir'}->dispose;
}


# pass parameters and archive info to harvester
sub harvest
{
   my ($self, $when, $what, $fromwhere) = @_;
   $self->SUPER::harvest ($self->{'archive'}, $when, $what, $fromwhere);   
}


# output log file header and start IR engine
sub processHarvestStart
{
   my ($self, $archiveidentifier, $archiveurl, $archiveinterval, 
       $archivemetadataPrefixes, $archiveset, $archivedaysoverlap,
       $archiveinterrequestgap) = @_;
   $self->{'log'}->log ("");
   $self->{'log'}->log ("Harvesting:");
   $self->{'log'}->log ("  archive         = $archiveidentifier");
   $self->{'log'}->log ("  url             = $archiveurl");
   $self->{'log'}->log ("  interval        = $archiveinterval days");
   foreach my $archivemetadataPrefix (@$archivemetadataPrefixes)
   {
      $self->{'log'}->log ("  metadataPrefix  = $archivemetadataPrefix");
   }
   $self->{'log'}->log ("  set             = $archiveset");
   $self->{'log'}->log ("  daysoverlap     = $archivedaysoverlap");
   $self->{'log'}->log ("  interrequestgap = $archiveinterrequestgap");
   $self->{'log'}->log ("");
   
   $self->{'ir'}->initializeBuffers;
}


# close down IR engine
sub processHarvestEnd
{
   my ($self) = @_;
   $self->{'ir'}->flushBuffers;
}


# remove old traces of an identifier from IR engine
sub processIdStart
{
   my ($self, $identifier, $status) = @_;
   
   $self->{'counter'}++;
   $self->{'log'}->log ("[$self->{counter}] Processing : $identifier");
   $self->{'ir'}->unindexIdentifier ($identifier);
}


# index all the terms in a record
sub processRecord
{
   my ($self, $xml, $date, $metadataPrefix, $set, $archiveid) = @_;

   # get identifier from header of record
   my $identifier = '';
   if ($xml->getElementsByTagName ('header', 0)->getLength == 1)
   {
      my $header = $xml->getElementsByTagName ('header', 0)->item(0);
      if ($header->getElementsByTagName ('identifier', 0)->getLength == 1)
      {
         my $identifiercontainer = $header->getElementsByTagName ('identifier', 0)->item(0);
         if (($identifiercontainer->hasChildNodes) && 
             ($identifiercontainer->getFirstChild->getNodeType == TEXT_NODE))
         {
            $identifier = $identifiercontainer->getFirstChild->getNodeValue;
         }
      }
   }
   $identifier =~ s/^([\s\r\n\t]*)([^\s\r\n\t]*)([\s\r\n\t]*)$/$2/;

   # get metadata pointer from record
   my $metadata = '';
   if ($xml->getElementsByTagName ('metadata', 0)->getLength == 1)
   {
      $metadata = $xml->getElementsByTagName ('metadata', 0)->item(0);
   }
   
   my $status = $xml->getAttribute ('status');
   
   if ($status eq 'deleted')
   {
      $self->{'log'}->log ("[$self->{counter}] Deleting: $identifier");
      $self->{'ir'}->unindexIdentifier ($identifier);
   }
   elsif (($identifier ne '') && ($metadata ne ''))
   {
      $self->{'log'}->log ("[$self->{counter}] Indexing: $identifier, $metadataPrefix");
      $self->{'ir'}->indexXML ($metadata, $identifier);
   }
}


# commit IR engine buffers
sub processIdEnd
{
   my ($self, $identifier) = @_;
   
   $self->{'log'}->log ("[$self->{counter}] Committing : $identifier");
   $self->{'ir'}->commitBuffer;
}


1;



=head1 NAME

Search - OAI Search Engine

=head1 SYNOPSIS

  use Search;
  
  my $s = new Search ('configfile.xml');
  
  $s->harvest;

  my $results = $s->search ('test');
  
  $s->close;

=head1 DESCRIPTION

Create a search engine based on an OAI-acessible data archive. This 
class builds upon I<Harvester> so all that information applies here 
as well.

The configuration file contains an XML description of the database to use
in addition to what is required for harvesting. For example,

  <?xml version="1.0" encoding="utf-8" ?>

  <searchconfig>
  
   <repositoryName>test search engine</repositoryName>
   <adminEmail>test@email.com</adminEmail>

   <database>DBI:mysql:irtest</database>
   <dbusername>auser</dbusername>
   <dbpassword>apassword</dbpassword>
   <table>search</table>
   
   <archive>
      <identifier>TEST</identifier>
      <url>http://www.test.org/cgi-bin/OAI.pl</url>
      <metadataPrefix>oai_dc</metadataPrefix>
      <interval>0.5</interval>
   </archive>
   
  </searchconfig>
  
The database specified must accept SQL queries, specifically 
select/insert/delete/create table/show tables.

=head1 METHODS

=over 4

=item new (config_filename)

Creates a new search engine object using the specified filename as 
a configuration.

=item harvest ()

Harvest metadata from all archives as specified in the configuration
file. 

=item close ()

Close database - call this after harvesting to dispose of the database
handle.

=item search (query, start, stop)

Execute the specified query on the search engine and return results
numbering from I<start> to I<stop> in the form of a reference to
a list.

=item searchrelated (identifier, start, stop)

Query for items related to the given identifier and return results
numbering from I<start> to I<stop> in the form of a reference to a 
list.

=back

=head1 NOTES

This is part of the larger effort to build componentized digital libraries.

=head1 CAVEATS

Of course this is untested :)

=head1 BUGS

None that I found yet.

=head1 AUTHOR

Hussein Suleman <hussein@vt.edu>

=head1 HISTORY

This is the first release version.

=cut

