#  ---------------------------------------------------------------------
#   General-purpose XML-based IR engine
#    v1.1
#    January 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------


package ODL::IRDB::XMLIR;


use Pure::EZXML;
use DBI;
use IR::Stemmer;
use IR::Stopper;


# constructor
sub new
{
   my ($classname, $database, $dbusername, $dbpassword, $dbindex, $rootxml) = @_;
   
   # cater for missing parameters
   if (! defined $classname)
   { die "This method has to be called as a constructor"; }
   if (! defined $database)
   { die "Missing database name"; }
   if (! defined $dbusername)
   { $dbusername = ''; }
   if (! defined $dbpassword)
   { $dbpassword = ''; }
   if (! defined $dbindex)
   { $dbindex = 'irindex'; }
   if (! defined $rootxml)
   { $rootxml = 'metadata'; }
   
   my $self = {
      database        => $database,
      dbusername      => $dbusername,
      dbpassword      => $dbpassword,
      
      dbindex         => $dbindex,
      
      stopper         => new IR::Stopper,
      stemmer         => new IR::Stemmer,
      
      rootxml         => $rootxml,
      feedback        => 10,
      
      wordbuffer      => {},
      tagbuffer       => {},
      dfbuffer        => {},
      identifierbuffer=> {},
      
      countbuffer     => {},
      bufferseparator => '!'
   };
   bless $self, $classname;
   
   $self->openDatabase;
   
   return $self;
}


# open database handle and create tables if necessary
sub openDatabase
{
   my ($self) = @_;
   
   $self->{dbh} = DBI->connect ($self->{database}, $self->{dbusername}, 
                                $self->{dbpassword}, { PrintError=>0 });
   
   # check for existance of index
   my $sth = $self->{dbh}->prepare ("select word from $self->{dbindex} limit 1");
   $sth->execute;
   my @row = $sth->fetchrow_array;
   $sth->finish;
   if ($#row > -1)
   {
      return;
   }

   # make new tables in database
   $sth = $self->{dbh}->do ("create table $self->{dbindex} (".
                            "word integer not null, ".
                            "tag smallint not null, ".
                            "identifier integer not null, ".
                            "count float not null".
                            ")");
   $sth = $self->{dbh}->do ("create index word on $self->{dbindex} (word)");
   $sth = $self->{dbh}->do ("create index tagwordcountidentifier on $self->{dbindex} (tag,word,count,identifier)");
   $sth = $self->{dbh}->do ("create index tagidentifiercount on $self->{dbindex} (tag,identifier,count)");

   $sth = $self->{dbh}->do ("create table $self->{dbindex}word (".
                            "id integer not null, ".
                            "word char(100) not null, ".
                            "primary key (id)".
                            ")");
   $sth = $self->{dbh}->do ("create index word on $self->{dbindex}word (word)");

   $sth = $self->{dbh}->do ("create table $self->{dbindex}tag (".
                            "id smallint not null, ".
                            "tag char(100) not null, ".
                            "primary key (id)".
                            ")");
   $sth = $self->{dbh}->do ("create index tag on $self->{dbindex}tag (tag)");
   
   $sth = $self->{dbh}->do ("create table $self->{dbindex}identifier (".
                            "id integer not null, ".
                            "identifier char(100) not null, ".
                            "primary key (id)".
                            ")");
   $sth = $self->{dbh}->do ("create index identifier on $self->{dbindex}identifier (identifier)");

   $sth = $self->{dbh}->do ("create table $self->{dbindex}df (".
                            "word integer not null, ".
                            "tag smallint not null, ".
                            "df float not null, ".
                            "primary key (word,tag)".
                            ")");
}


# destructor
sub dispose
{
   my ($self) = @_;
   $self->{dbh}->disconnect;
}


# get current counters and set up index buffers
sub initializeBuffers
{
   my ($self) = @_;
   
   my $sth;
   
   $sth = $self->{dbh}->prepare ("select count(*) from $self->{dbindex}word");
   $sth->execute;
   $self->{nextword} = ($sth->fetchrow_array)[0] + 1;
   $sth->finish;

   $sth = $self->{dbh}->prepare ("select count(*) from $self->{dbindex}tag");
   $sth->execute;
   $self->{nexttag} = ($sth->fetchrow_array)[0] + 1;
   $sth->finish;
   
   $sth = $self->{dbh}->prepare ("select count(*) from $self->{dbindex}identifier");
   $sth->execute;
   $self->{nextidentifier} = ($sth->fetchrow_array)[0] + 1;
   $sth->finish;
   
   $self->{wordbuffer} = {};
   $self->{tagbuffer} = {};
   $self->{dfbuffer} = {};
   $self->{identifierbuffer} = {};
}


# store buffer on disk
sub flushBuffer
{
   my ($self, $bufferptr, $tablename) = @_;
   
   foreach my $key (keys %$bufferptr)
   {
      $self->{dbh}->do ("delete from $self->{dbindex}$tablename where id=\'$bufferptr->{$key}\'");
      $self->{dbh}->do ("insert into $self->{dbindex}$tablename values ($bufferptr->{$key}, \'$key\')");
   }

   $bufferptr = {};
}


# store df buffer on disk
sub flushdfBuffer
{
   my ($self) = @_;
   
   foreach my $key (keys %{$self->{dfbuffer}})
   {
      my @keypart = split ($self->{bufferseparator}, $key);
      $self->{dbh}->do ("delete from $self->{dbindex}df where word=\'$keypart[0]\' and tag=\'$keypart[1]\'");
      $self->{dbh}->do ("insert into $self->{dbindex}df values ($keypart[0], $keypart[1], $self->{dfbuffer}->{$key})");
   }
   
   $self->{dfbuffer} = {};
}


# flush all buffers
sub flushBuffers
{
   my ($self) = @_;
   
   $self->flushBuffer ($self->{wordbuffer}, 'word');
   $self->flushBuffer ($self->{tagbuffer}, 'tag');
   $self->flushBuffer ($self->{identifierbuffer}, 'identifier');
   $self->flushdfBuffer;
}


# create a hash of words with counts from a string
sub makeWordList
{
   my ($self, $str) = @_;
   
   my @words = split (' ', $str);
   my %wordcounts = ();
   
   # create list of words and counts
   foreach my $aword (@words)
   {
      $aword = lc ($aword);
      $aword = $self->{stopper}->stop ($aword);
      
      if ($aword ne '')
      {
         $aword = $self->{stemmer}->stem ($aword);
      }
      
      if ($aword ne '')
      {
         if (! exists $wordcounts{$aword})
         { 
            $wordcounts{$aword} = 0; 
         }
         $wordcounts{$aword} += 1;
      }
   }
   
   \%wordcounts;   
}


# index a list of words
sub indexWordList
{
   my ($self, $wordlist, $tag, $identifier) = @_;
   
   # calculate document length   
   my $documentlength = 0;
   foreach my $aword (keys %$wordlist)
   {
      $documentlength += $wordlist->{$aword};
   }

   # add normalized entries to the index
   foreach my $aword (keys %$wordlist)
   {
      my $normalizedcount = $wordlist->{$aword} / $documentlength;
      
      my $bufferkey = join ($self->{bufferseparator}, ($aword, $tag, $identifier));
      if (defined $self->{countbuffer}->{$bufferkey})
      {
         my ($count, $mdsets) = split ($self->{bufferseparator}, $self->{countbuffer}->{$bufferkey});
         $normalizedcount += $count * $mdsets;
         $mdsets += 1;
         $normalizedcount /= $mdsets;
         my $newentry = join ($self->{bufferseparator}, ($normalizedcount, $mdsets));
         $self->{countbuffer}->{$bufferkey} = $newentry;
      }
      else
      {
         my $newentry = join ($self->{bufferseparator}, ($normalizedcount, 1));
         $self->{countbuffer}->{$bufferkey} = $newentry;
      }
   }
}


# write out all buffers to the databases
sub commitBuffer
{
   my ($self) = @_;
   
   my $valuebuffer = '';
   
   foreach my $bufferkey (keys %{$self->{countbuffer}})
   {
      my ($word, $tag, $identifier) = split ($self->{bufferseparator}, $bufferkey);
      my ($count, $mdsets) = split ($self->{bufferseparator}, $self->{countbuffer}->{$bufferkey});

      my $sth;
      my ($wordid, $tagid, $identifierid);
      my @row;
      
      # word table
      if (! defined $self->{wordbuffer}->{$word})
      {
         $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}word where word=\'$word\'");
         $sth->execute;
         if (@row = $sth->fetchrow_array)
         {
            $self->{wordbuffer}->{$word} = $row[0];
         }
         else
         {
            $self->{wordbuffer}->{$word} = $self->{nextword}++;
         }
         $sth->finish;
      }
      $wordid = $self->{wordbuffer}->{$word};
      
      # tag table
      if (! defined $self->{tagbuffer}->{$tag})
      {
         $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}tag where tag=\'$tag\'");
         $sth->execute;
         if (@row = $sth->fetchrow_array)
         {
            $self->{tagbuffer}->{$tag} = $row[0];
         }
         else
         {
            $self->{tagbuffer}->{$tag} = $self->{nexttag}++;
         }
         $sth->finish;
      }
      $tagid = $self->{tagbuffer}->{$tag};
      
      # identifier table
      if (! defined $self->{identifierbuffer}->{$identifier})
      {
         $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}identifier where identifier=\'$identifier\'");
         $sth->execute;
         if (@row = $sth->fetchrow_array)
         {
            $self->{identifierbuffer}->{$identifier} = $row[0];
         }
         else
         {
            $self->{identifierbuffer}->{$identifier} = $self->{nextidentifier}++;
         }
         $sth->finish;
      }
      $identifierid = $self->{identifierbuffer}->{$identifier};

      # df table
      my $df = join ($self->{bufferseparator}, ($wordid, $tagid));
      if (! defined $self->{dfbuffer}->{$df})
      {
         $sth = $self->{dbh}->prepare ("select df from $self->{dbindex}df where word=\'$wordid\' and tag=\'$tagid\'");
         $sth->execute;
         if (@row = $sth->fetchrow_array)
         {
            $self->{dfbuffer}->{$df} = $row[0];
         }
         else
         {
            $self->{dfbuffer}->{$df} = 0;
         }
         $sth->finish;
      }
      $self->{dfbuffer}->{$df} += $count;
      
# optimization for mysql
#      $self->{dbh}->do ("insert into $self->{dbindex} values ($wordid, $tagid, $identifierid, $count)");

# begin optimization segment 1
      if ($valuebuffer ne '')
      {
         $valuebuffer .= ',';
      }
      $valuebuffer .= "($wordid, $tagid, $identifierid, $count)";
      if (length ($valuebuffer) > 10000)
      {
         my $sth = $self->{dbh}->prepare ("insert into $self->{dbindex} values $valuebuffer");
         $sth->execute;
         $sth->finish;
         $valuebuffer = '';
      }
# end optimization segment 1

   }
   
# begin optimization segment 2
   if (length ($valuebuffer) > 0)
   {
      my $sth = $self->{dbh}->prepare ("insert into $self->{dbindex} values $valuebuffer");
      $sth->execute;
      $sth->finish;
   }
# end optimization segment 2
   
   %{$self->{countbuffer}} = ();
}


# remove indexing information for all words associated with an id
sub unindexIdentifier
{
   my ($self, $identifier) = @_;
   
   my $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}identifier where identifier=\'$identifier\'");
   $sth->execute;
   if (my @row = $sth->fetchrow_array)
   {
      $sth->finish;
      $sth = $self->{dbh}->prepare ("delete from $self->{dbindex} where identifier=\'$row[0]\'");
      $sth->execute;
   }
   $sth->finish;
}


# remove all whitespace and empty nodes from tree
sub cleanWhitespace
{
   my ($self, $elem) = @_;
   
   if ($elem->getNodeName eq '#text')
   {
      my $str = $elem->getNodeValue;

      # remove leading whitespace
      $str =~ s/^([\s\r\n\t]*)(.*)/$2/o;
      # remove trailing whitespace
      $str =~ s/(.*)([\s\r\n\t]*)$/$1/o;
      # remove multiple spaces, interleaved whitespace, punctuation
      $str =~ s/[\s\r\n\t,.;:'"=\+\\\/\(\)\[\]\{\}\-0-9]+/ /g;
      
      $elem->setNodeValue ($str);
   }
   
   foreach my $kid ($elem->getChildNodes)
   {
      $self->cleanWhitespace ($kid);
      if (($kid->getNodeName () eq '#text') && ($kid->getNodeValue () eq ''))
      { $elem->removeChild ($kid); }
   }
}


# index all nodes in a tree
sub processNodes
{
   my ($self, $node, $identifier) = @_;
   
   my %wordlist = ();
   
   if ($node->getNodeType == TEXT_NODE)
   {
      my $newwordlist = $self->makeWordList ($node->getNodeValue);
      %wordlist = %$newwordlist;
   }
   elsif ($node->getNodeType () == ELEMENT_NODE)
   {
      foreach my $kid ($node->getChildNodes)
      {
         my $newwordlist = $self->processNodes ($kid, $identifier);
         foreach my $aword (keys %$newwordlist)
         {
            if (! exists $wordlist{$aword})
            {
               $wordlist{$aword} = 0;
            }
            $wordlist{$aword} += $newwordlist->{$aword};
         }
      }
      $self->indexWordList (\%wordlist, $node->getNodeName, $identifier);
   }
   
   \%wordlist;
}


# index an XML fragment
sub indexXML
{
   my ($self, $xml, $identifier) = @_;
   
   $self->cleanWhitespace ($xml);
   $self->processNodes ($xml, $identifier);
}


# execute a search operation on the index databases
sub search
{
   my ($self, $query, $start, $stop, $trans) = @_;

   # check for NULL query
   if ((! defined $query) || ($query eq ''))
   { return []; }

   my %positives = ();
   my %negatives = ();
   my %neutrals = ();
   my $gotpositives = 0;
   my @terms = split (' ', $query);

   # iterate over all terms
   foreach my $aterm (@terms)
   {
      # check for positive/negative prefix
      my $sign = '';
      if (substr ($aterm, 0, 1) eq '-')
      { 
         $aterm = substr ($aterm, 1);
         $sign = '-';
      }
      elsif (substr ($aterm, 0, 1) eq '+')
      {
         $aterm = substr ($aterm, 1);
         $sign = '+';
      }
      
      # check for category prefix
      my @parts = split (':', $aterm);
      my $word = $parts[$#parts];
      my $tag = $self->{rootxml};
      if ($#parts > 0)
      {
         $tag = $parts[0];
      }
      
      # check for weight suffix
      @parts = split (';', $word);
      my $weight = 1;
      if ($#parts > 0)
      {
         $word = $parts[0];
         $weight = $parts[$#parts];         
      }
      
      # check for stopwords
      $word = $self->{stopper}->stop ($word);
      
      # perform stemming on query term
      if ($word ne '')
      {
         $word = $self->{stemmer}->stem ($word);
      }
      
      # search for term in inverted file
      if ($word ne '')
      {
         my $sth;
         my ($wordid, $tagid);
         my @row;
         
         $tagid = 0;
         $wordid = 0;
      
         if (defined $trans)
         {
            $tagid = $tag;
            $wordid = $word;
         }
         else
         {
            # get word id
            $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}word where word=\'$word\'");
            $sth->execute;
            if (@row = $sth->fetchrow_array)
            {
               $wordid = $row[0];
            }
            $sth->finish;
      
            # get tag id
            $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}tag where tag=\'$tag\'");
            $sth->execute;
            if (@row = $sth->fetchrow_array)
            {
               $tagid = $row[0];
            }
            $sth->finish;
         }
         
         if (($tagid > 0) && ($wordid > 0))
         {
            # get df
            my $df = 1;
            $sth = $self->{dbh}->prepare ("select df from $self->{dbindex}df where word=\'$wordid\' and tag=\'$tagid\'");
            $sth->execute;
            if (@row = $sth->fetchrow_array)
            {
               $df = $row[0];
            }
         
            # get results from database
            my @partialresults = ();
            my @partialcounts = ();
            my ($prr, $pcc);
            my $stopvalue = '';
            if (($stop > 0) && ($sign ne '-'))
            {
               $stop += 20;
               $stopvalue = "order by count desc limit $stop";
               $stop -= 20;
            }
            $sth = $self->{dbh}->prepare ("select identifier, count from $self->{dbindex} where word=\'$wordid\' and tag=\'$tagid\' $stopvalue");
            $sth->execute;
            $sth->bind_col (1, \$prr);
            $sth->bind_col (2, \$pcc);
            while ($sth->fetchrow)
            {
               push (@partialresults, $prr);
               push (@partialcounts, $pcc);
            }
            $sth->finish;

            # unify results with negatives list
            if ($sign eq '-')
            {
               foreach my $result (@partialresults)
               {
                  $negatives{$result} = 1;
               }
            }
      
            # intersect result with positives list
            elsif ($sign eq '+')
            {
               my %partialpositive = ();
               foreach my $result (@partialresults)
               {
                  $partialpositive{$result} = 1;
               }
               if ($gotpositives == 0)
               {
                  $gotpositives = 1;
                  %positives = %partialpositive;
               }
               else
               {
                  foreach my $akey (keys %positives)
                  {
                     if (!(exists $partialpositive{$akey}))
                     { 
                        delete $positives{$akey};
                     }
                  }
               }
            }
      
            # add all non-negatives to final result list, 
            # calculating weights by tf.idf
            if ($sign ne '-')
            {
               for (my $i=0; $i<=$#partialresults; $i++ )
               {
                  my $result = $partialresults[$i];
                  my $count = $partialcounts[$i] / $df;
                  if (!(exists $neutrals{$result}))
                  {
                     $neutrals{$result} = $count * $weight;
                  }
                  else
                  {
                     $neutrals{$result} += $count * $weight;
                  }
               }
            }
         }
      }
   }

   # check for positives and remove negatives from result list
   foreach my $result (keys %neutrals)
   {
      if ((exists $negatives{$result}) ||
          (($gotpositives == 1) && (!(exists $positives{$result}))))
      {
         delete $neutrals{$result};
      }
   }
   
   # sort result list
   my @finallist = ();
   foreach my $akey (keys %neutrals)
   {
      $finallist[$#finallist+1] = [ $akey, $neutrals{$akey} ];
   }
   @finallist = sort { $b->[1] <=> $a->[1] } @finallist;
   
   # get actual identifiers
   my @resultlist = ();
   if ($start == 0) { $start = 1; }
   if ($stop == 0) { $stop = $#finallist+1; }
   for ( my $i=$start; $i<=$stop; $i++ )
   {
      if (($i-1) <= $#finallist)
      {
         my $finalid = $finallist[$i-1]->[0];
   
         # get identifier id
         my $sth = $self->{dbh}->prepare ("select identifier from $self->{dbindex}identifier where id=\'$finalid\'");
         $sth->execute;
         if (my @row = $sth->fetchrow_array)
         {
            while (substr ($row[0], -1) eq ' ') { chop $row[0]; };
            push (@resultlist, $row[0]);
         }
      
         $sth->finish;
      }
   }
   
   push (@resultlist, $#finallist+1);
   
   \@resultlist;
}


# execute a search for items related to a given identifier
sub searchrelated
{
   my ($self, $identifier, $start, $stop) = @_;
   
   # check for empty query
   if ((! defined $identifier) || ($identifier eq ''))
   { return []; }

   # get identifier id
   my $identifierid = 0;
   my $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}identifier where identifier=\'$identifier\'");
   $sth->execute;
   if (my @row = $sth->fetchrow_array)
   {
      $identifierid = $row[0];
   }
   $sth->finish;

   # get tag id
   my $tagid = 0;
   $sth = $self->{dbh}->prepare ("select id from $self->{dbindex}tag where tag=\'$self->{rootxml}\'");
   $sth->execute;
   if (my @row = $sth->fetchrow_array)
   {
      $tagid = $row[0];
   }
   $sth->finish;

   # get highest count terms
   $sth = $self->{dbh}->prepare ("select word, tag, count from $self->{dbindex} where identifier=\'$identifierid\' and tag=\'$tagid\' order by count desc limit $self->{feedback}");
   $sth->execute;

   # build query
   my $rowref;
   my $query = '';
   while ($rowref = $sth->fetchrow_hashref)
   {
      $query .= $rowref->{tag}.':'.$rowref->{word}.';'.$rowref->{count}.' ';
   }
   $sth->finish;

   # perform query
   my $resultlist = $self->search ($query, $start, $stop, 1);
   
   $resultlist;   
}


1;


=head1 NAME

XMLIR - Information Retrieval on XML data using XML::DOM parser

=head1 SYNOPSIS

  use XMLIR;
 
  # create new IR object
  my $ir = new ODL::IRDB::XMLIR ('DBI:mysql:irtest', 'username', 'userpwd', 'index', 'metadata');
  
  # index a single word without stemming/stopping
  $ir->indexWord ('aword', 'acategory', 'anidentifier');
  
  # index a sentence
  $ir->indexSentence ('the big brown fox', ['cat1', 'cat2'], 'anidentifier');
  
  # index an xml file
  my $parser = new Pure::EZXML;
  my $doc = $parser->parsefile ("file.xml");
  $ir->indexXML ($doc->getDocumentElement, 'dummydocid');
  
  # remove an entry from index
  $ir->unindexIdentifier ('anidentifier');

  # search through index
  my $results;
  $results = $ir->search ('big brown');
  $results = $ir->search ('big +brown');
  $results = $ir->search ('big cat1:brown');
  $results = $ir->search ('cat1:big cat2:brown');  
  $results = $ir->search ('big -brown');  
  $results = $ir->search ('+big -brown');
  foreach (@$results) { print; }
  
  # search through index for related items
  my $results;
  $results = $ir->searchrelated ('oai:ETDS:12345');

=head1 DESCRIPTION

Information Retrieval using a vector model completely in
native Perl code.

The index is stored in a mysql database, with the following fields

=over 4

=item word varchar(100) not null

=item tag varchar(100) not null

=item identifier varchar(100) not null

=item count int not null

=back

Search is implemented as a pre-normalized tf.idf vector summed
over all componenets. Search terms are given as a string of 
space-separated terms, with the following format:

=over 4

=item [+/-] [<category>:] <term>

=back

First all results for positive search terms are ranked. Then,
if there are mandatory terms (prefixed by '+'), those entries are 
selected and the rest are discarded. Also, if there are excluded 
terms (prefixed by '-'), any entries with those are also discarded.

X<information retrieval>
X<search engine>

=head1 METHODS

=over 4

=item new (database, dbuser, dbpassword, dbindex, rootxml)

Constructor to create new IR object. I<database> is the DBI
string that is used to access a database, for example 
'DBI:mysql:irtest'. I<dbuser> and I<dbpassword> are the
username/password used to access the database - this must
permit select/insert/createtable/update operations. I<dbindex>
is the name of the table that stores the index data - this can
be anything at all, but by allowing it to be configurable,
it is possible to store multiple indices in one database.
I<rootxml> is the root element of all xml submitted to the IR
system. This must be consistent since it is assumed that all XML 
elements have a common root for the purposes of efficiently
indexing and searching sub-categories - if necessary, this
should be added synthetically to the XML tree before submitting
it for indexing.

=item indexXML (XML::DOM::Node, identifier)

Traverse the DOM tree and index all text nodes at their current
encalsulating nodes as well as all their ancestors, using stemming
and stopping.

=item search (query)

Search through the index as described above.

=item searchrelated (identifier)

Search through the index for documents similar to the specified one.

=item unindexIdentifier (identifier)

Removes all entries for the given identifier from the index.

=back

=head1 NOTES

This is part of the larger effort to build a trivial search engine
component for componentized digital libraries.

=head1 CAVEATS

Of course this is untested :)

=head1 BUGS

None that I found yet.

=head1 AUTHOR

Hussein Suleman <hussein@vt.edu>

=head1 HISTORY

This is the first release version.

=cut

