#  ---------------------------------------------------------------------
#   General database implementation of Suggest (v1.0) - data provider
#    v1.0
#    April 2002
#  ------------------+--------------------+-----------------------------
#   Hussein Suleman  |   hussein@vt.edu   |    www.husseinsspace.com    
#  ------------------+--------------------+-+---------------------------
#   Department of Computer Science          |        www.cs.vt.edu       
#     Digital Library Research Laboratory   |       www.dlib.vt.edu      
#  -----------------------------------------+-------------+-------------
#   Virginia Polytechnic Institute and State University   |  www.vt.edu  
#  -------------------------------------------------------+-------------


package ODL::Suggest::SuggestDP;


use POSIX;
use DBI;

use Pure::EZXML;
use Pure::X2D;
use ODL::Suggest::SuggestDB;

use XOAI::XOAIDP;
use vars ('@ISA');
@ISA = ("XOAI::XOAIDP");


# constructor
sub new
{
   my ($classname, $configfile) = @_;
   
   my $self = $classname->SUPER::new ();

   # get configuration from file
   my $con = new Pure::X2D ($configfile);
   $self->{'database'} = $con->param ('database', '');
   $self->{'dbusername'} = $con->param ('dbusername', '');
   $self->{'dbpassword'} = $con->param ('dbpassword', '');
   $self->{'table'} = $con->param ('table', 'odlsuggest');
   $self->{'repositoryName'} = $con->param ('repositoryName', 'ODL Suggest Engine');
   $self->{'adminEmail'} = $con->param ('adminEmail', "mailto:someone\@somewhere");
   $self->{'recordlimit'} = $con->param ('recordlimit', 100);
   $self->{'archiveId'} = $con->param ('archiveId', 'box');
   $self->{'maxrecent'} = $con->param ('maxrecent', 5);
   $self->{'maxvote'} = $con->param ('maxvote', 5);
   $self->{'maxcount'} = $con->param ('maxcount', 100);
  
   # read in access control lists
   $self->{'allowread'} = {};
   foreach my $machine ($con->param ('allowread', []))
   {
      my ($a, $b, $c, $d) = unpack ('C4', (gethostbyname ($machine))[4]);
      my $ipaddress = "$a.$b.$c.$d";
      $self->{'allowread'}->{$ipaddress} = 1;
   }
   $self->{'allowwrite'} = {};
   foreach my $machine ($con->param ('allowwrite', []))
   {
      my ($a, $b, $c, $d) = unpack ('C4', (gethostbyname ($machine))[4]);
      my $ipaddress = "$a.$b.$c.$d";
      $self->{'allowwrite'}->{$ipaddress} = 1;
   }
   
   # set internal variables
   $self->{'resumptionseparator'} = '!';
   
   # open database
   $self->{'dbh'} = ODL::Suggest::SuggestDB::openDatabase ($self->{'database'},
     $self->{'dbusername'}, $self->{'dbpassword'}, $self->{'table'});

   $self->{'odlsearch_xmlnsprefix'} = 'http://oai.dlib.vt.edu/ODL/Search/1.0';

   bless $self, $classname;
   return $self;
}


# destructor
sub dispose
{
   my ($self) = @_;
   $self->SUPER::dispose ();
   $self->{'dbh'}->disconnect;
}


# convert unix calendar time to ISO8601 format
sub toISO8601
{
   my ($self, $atime, $granularity) = @_;
   
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime ($atime);
   if ((defined $granularity) && ($granularity eq 'second'))
   {
      sprintf ("%04d-%02d-%02dT%02d:%02d:%02d+00:00", $year+1900, $mon+1, $mday, $hour, $min, $sec);
   }
   else
   {
      sprintf ("%04d-%02d-%02d", $year+1900, $mon+1, $mday);
   }
}


# convert an ISO8601 date into a unix calendar time representation
sub toSeconds
{
   my ($self, $date, $from) = @_;
   
   my ($hour, $minute, $second);
   
   if ((defined $from) && ($from == 1))
   {
      ($hour, $minute, $second) = (0, 0, 0);
   }
   else
   {
      ($hour, $minute, $second) = (23, 59, 59);
   }
   
   if ($date =~ /([0-9]{4})-([0-9]{2})-([0-9]{2})T([0-9]{2}):([0-9]{2}):([0-9]{2})[\+\-][0-9]{2}:[0-9]{2}/)
   {
      return mktime ($6, $5, $4-1, $3, $2-1, $1-1900);
   }
   elsif ($date =~ /([0-9]{4})-([0-9]{2})-([0-9]{2})/)
   {
      return mktime ($second, $minute, $hour, $3, $2-1, $1-1900);
   }
   return 0;
}


# component-specific identification
sub Archive_Identify
{
   my ($self) = @_;
   
   my $id = $self->SUPER::Archive_Identify ();
      
   if (! exists $id->{'description'})
   {
      $id->{'description'} = [];
   }
   my $desc = {
      'odl-description' => [[ 
         {
            'xmlns' => $self->{xoai_xmlnsprefix}.'/odl-description',
            'xsi:schemaLocation' => 
                       $self->{xoai_xmlnsprefix}.'/odl-description '.
                       $self->{xoai_xmlnsprefix}.'/odl-description.xsd'
         },
         {
            'protocol' => 'ODL-Recommend',
            'version'  => '1.0',
            'mdorder'  => [ qw ( protocol version ) ]
         } 
      ]]
   };
   push (@{$id->{'description'}}, $desc);
   
   $id;
}


# retrieve records from the source archive as required
sub Archive_FormatRecord
{
   my ($self, $hashref, $metadataFormat) = @_;
   
   if (defined $hashref->{$metadataFormat})
   {
      my $rowref = $hashref->{$metadataFormat};
      $self->FormatRecord ($rowref->{'identifier'},
                           $self->toISO8601 ($rowref->{'datestamp'}, 'second'),
                           $rowref->{'metadata'},
                           $rowref->{'about'},
                           $rowref->{'status'}
                          );
   }
   elsif ((keys %$hashref) == 0)
   {
      return '';
   }
   else
   {
      my $firstkey = (keys %$hashref)[0];
      my $rowref = $hashref->{$firstkey};
      $self->FormatRecord ($rowref->{'identifier'},
                           $self->toISO8601 ($rowref->{'datestamp'}, 'second'),
                           '',
                           '',
                           $rowref->{'status'}
                          );
   }
}


# pass request through to source
sub Archive_GetRecord
{
   my ($self, $identifier, $metadataPrefix) = @_;
   
   my %records = ();
   
   if ($metadataPrefix eq 'oai_dc')
   {
      my $id = '';
      if ($identifier =~ /[^:]+:[^:]+:(.*)/)
      {
         $id = $1;
      }
      
      my $sth = $self->{'dbh'}->prepare ("select * from $self->{'table'} where seq=\'$id\'");
      $sth->execute;

      while (my $rowref = $sth->fetchrow_hashref)
      {
         my $metadata = "<dc xmlns=\"http://purl.org/dc/elements/1.1/\" xsi:schemaLocation=\"http://purl.org/dc/elements/1.1/ http://www.openarchives.org/OAI/1.1/dc.xsd\">\n\n".
            "<identifier>$rowref->{'subject'}</identifier>\n".
            "<relation>$rowref->{'object'}</relation>\n".
            "</dc>\n";

         $records{'oai_dc'} = {
            identifier => $rowref->{'seq'},
            datestamp  => $rowref->{'datestamp'},
            status     => '',
            metadata   => $metadata,
            about      => ''
         };
      }
      
      $sth->finish;
   }

   \%records;
}


# return list of records/identifiers
sub Archive_ListRecords
{
   my ($self, $set, $from, $until, $metadataPrefix, $resumptionToken) = @_;

   my @allrows = ();
   my $hits = 0;
   
   if ($set eq '')
   {
      my ($offset);
      if ($resumptionToken eq '')
      {
         $offset = 0;
      }
      else
      {
         my @rdata = split ($self->{resumptionseparator}, $resumptionToken);
         ($set, $from, $until, $metadataPrefix, $offset) = @rdata;
      }

      my $sth;

      # construct query
      my $sqlquery = "select * from $self->{'table'}";

      if ($from eq '')
      {
         if ($until eq '') 
         { }
         else
         { $sqlquery = $sqlquery." where (datestamp <= \'".$self->toSeconds ($until)."\')"; }
      }
      else
      { 
         if ($until eq '')
         { $sqlquery = $sqlquery." where (datestamp >= \'".$self->toSeconds ($from, 1)."\')"; }
         else
         { $sqlquery = $sqlquery." where (datestamp >= \'".$self->toSeconds ($from, 1)."\') and (datestamp <= \'".$self->toSeconds ($until)."\')"; }
      }
      
      my $limit = $self->{'recordlimit'} + 1;
      $sqlquery = $sqlquery." order by seq desc limit $offset,$limit";
      $sth = $self->{'dbh'}->prepare ($sqlquery);
      $sth->execute;

      my $gotmore = 0;
      my $count = 0; 

      while (my $rowref = $sth->fetchrow_hashref)
      {
         if ($count == $self->{'recordlimit'})
         { $gotmore = 1; }
         else
         {
            $count++;
            if ($metadataPrefix eq '')
            {
               push (@allrows, { 
                  urn    => $rowref->{'seq'},
                  status => ''
               });
            }
            else
            {
               my $metadata = "<dc xmlns=\"http://purl.org/dc/elements/1.1/\" xsi:schemaLocation=\"http://purl.org/dc/elements/1.1/ http://www.openarchives.org/OAI/1.1/dc.xsd\">\n\n".
                  "<identifier>$rowref->{'subject'}</identifier>\n".
                  "<relation>$rowref->{'object'}</relation>\n".
                  "</dc>\n";

               push (@allrows, {
                  $metadataPrefix => {
                     identifier => $rowref->{'seq'},
                     datestamp  => $rowref->{'datestamp'},
                     status     => '',
                     metadata   => $metadata,
                     about      => ''
                  }
               });
            }
         }
      }

      $sth->finish;

      $resumptionToken = '';
      if ($gotmore == 1)
      {
         $resumptionToken = join ($self->{'resumptionseparator'}, ($set,$from,$until,$metadataPrefix,$offset+$self->{'recordlimit'}));
      }
   }
   elsif ($set =~ /([rRpP]{2})\/([0-9]+)\/([0-9]+)\/(.*)/)
   {
      my $mode = uc ($1);
      my $start = $2 - 1;
      my $stop = $3 - $2 + 1;
      my $id = $4;
      
      # get people that accessed an object
      if ($mode eq 'RP')
      {
         # get entries
         my $sth = $self->{'dbh'}->prepare ("select subject from $self->{'table'} where object=\'$id\' order by seq desc limit $start,$stop");
         $sth->execute;
         while (my $rowref = $sth->fetchrow_hashref)
         {
            push (@allrows, {
               urn    => $rowref->{'subject'},
               status => ''
            });
         }
         $sth->finish;
         
         # get count
         $sth = $self->{'dbh'}->prepare ("select count(*) from $self->{'table'} where object=\'$id\'");
         $sth->execute;
         while (my @rowref = $sth->fetchrow_array)
         {
            $hits = $rowref[0];
         }
         $sth->finish;
      }
      # get other objects accessed by people who accessed an object
      elsif ($mode eq 'RR')
      {
         my $count = 0;
         my $vote = 0;
         my %fulllist = ();
         
         # get list of people who accessed an object
         my $sth = $self->{'dbh'}->prepare ("select subject from $self->{'table'} where object=\'$id\' order by seq desc");
         $sth->execute;
         while ((my $rowref = $sth->fetchrow_hashref) && 
                ($count < $self->{'maxcount'}) && 
                ($vote < $self->{'maxvote'}))
         {
            # get list of objects access by each person and add to list
            my $subject = $rowref->{'subject'};
            my $sth2 = $self->{'dbh'}->prepare ("select object from $self->{'table'} where subject=\'$subject\' and object!=\'$id\' order by seq desc limit $self->{'maxrecent'}");
            $sth2->execute;
            while (my $rowref2 = $sth2->fetchrow_hashref)
            {
               if (! exists $fulllist{$rowref2->{'object'}})
               {
                  $fulllist{$rowref2->{'object'}} = 0;
                  $count++;
               }
               $fulllist{$rowref2->{'object'}}++;
               if ($fulllist{$rowref2->{'object'}} > $vote)
               {
                  $vote = $fulllist{$rowref2->{'object'}};
               }
            }
            $sth2->finish;
         }
         $sth->finish;
         
         # sort list and gather top entries
         my @splitlist = map { [ $_, $fulllist{$_} ] } keys %fulllist;
         my @sortlist = sort { $b->[1] <=> $a->[1] } @splitlist;
         for ( my $i=$start; ($i<($start+$stop) && ($i<=$#sortlist)); $i++ )
         {
            push (@allrows, {
               urn    => $sortlist[$i]->[0],
               status => ''
            });
         }
         $hits = $#sortlist + 1;
      }
      # get other people who accessed objects accessed by a person
      elsif ($mode eq 'PP')
      {
         my $count = 0;
         my $vote = 0;
         my %fulllist = ();
         
         # get list of objects accessed by the person
         my $sth = $self->{'dbh'}->prepare ("select object from $self->{'table'} where subject=\'$id\' order by seq desc");
         $sth->execute;
         while ((my $rowref = $sth->fetchrow_hashref) && 
                ($count < $self->{'maxcount'}) && 
                ($vote < $self->{'maxvote'}))
         {
            # get list of people who accessed each object and add to list
            my $object = $rowref->{'object'};
            my $sth2 = $self->{'dbh'}->prepare ("select subject from $self->{'table'} where object=\'$object\' and subject!=\'$id\' order by seq desc limit $self->{'maxrecent'}");
            $sth2->execute;
            while (my $rowref2 = $sth2->fetchrow_hashref)
            {
               if (! exists $fulllist{$rowref2->{'subject'}})
               {
                  $fulllist{$rowref2->{'subject'}} = 0;
                  $count++;
               }
               $fulllist{$rowref2->{'subject'}}++;
               if ($fulllist{$rowref2->{'subject'}} > $vote)
               {
                  $vote = $fulllist{$rowref2->{'subject'}};
               }
            }
            $sth2->finish;
         }
         $sth->finish;
         
         # sort list and gather top entries
         my @splitlist = map { [ $_, $fulllist{$_} ] } keys %fulllist;
         my @sortlist = sort { $b->[1] <=> $a->[1] } @splitlist;
         for ( my $i=$start; ($i<($start+$stop) && ($i<$#sortlist)); $i++ )
         {
            push (@allrows, {
               urn    => $sortlist[$i]->[0],
               status => ''
            });
         }
         $hits = $#sortlist + 1;
      }
      # get objects accessed by other people who accessed objects accessed by a person
      elsif ($mode eq 'PR')
      {
         my $count = 0;
         my $vote = 0;
         my %fulllist = ();
         
         # get list of objects accessed by person
         my $sth = $self->{'dbh'}->prepare ("select object from $self->{'table'} where subject=\'$id\' order by seq desc");
         $sth->execute;
         while ((my $rowref = $sth->fetchrow_hashref) && 
                ($count < $self->{'maxcount'}) && 
                ($vote < $self->{'maxvote'}))
         {
            # get list of people who accessed each object and add to list
            my $object = $rowref->{'object'};
            my $sth2 = $self->{'dbh'}->prepare ("select subject from $self->{'table'} where object=\'$object\' and subject!=\'$id\' order by seq desc limit $self->{'maxrecent'}");
            $sth2->execute;
            while (my $rowref2 = $sth2->fetchrow_hashref)
            {
               if (! exists $fulllist{$rowref2->{'subject'}})
               {
                  $fulllist{$rowref2->{'subject'}} = 0;
                  $count++;
               }
               $fulllist{$rowref2->{'subject'}}++;
               if ($fulllist{$rowref2->{'subject'}} > $vote)
               {
                  $vote = $fulllist{$rowref2->{'subject'}};
               }
            }
            $sth2->finish;
         }
         $sth->finish;
         
         # sort list to get most best relationships at head of list
         my @splitlist = map { [ $_, $fulllist{$_} ] } keys %fulllist;
         my @sortlist = sort { $b->[1] <=> $a->[1] } @splitlist;
         
         $count = 0;
         $vote = 0;
         %fulllist = ();
         my $ptr = 0;
         
         while (($ptr <= $#sortlist) && 
                ($count < $self->{'maxcount'}) && 
                ($vote < $self->{'maxvote'}))
         {
            # get list of objects for each person in sorted person list and add to object list
            my $subject = $sortlist[$ptr]->[0];
            my $sth2 = $self->{'dbh'}->prepare ("select object from $self->{'table'} where subject=\'$subject\' order by seq desc limit $self->{'maxrecent'}");
            $sth2->execute;
            while (my $rowref2 = $sth2->fetchrow_hashref)
            {
               if (! exists $fulllist{$rowref2->{'object'}})
               {
                  $fulllist{$rowref2->{'object'}} = 0;
                  $count++;
               }
               $fulllist{$rowref2->{'object'}}++;
               if ($fulllist{$rowref2->{'object'}} > $vote)
               {
                  $vote = $fulllist{$rowref2->{'object'}};
               }
            }
            $sth2->finish;
            $ptr++;
         }
         
         # sort object list and gather top entries
         @splitlist = map { [ $_, $fulllist{$_} ] } keys %fulllist;
         @sortlist = sort { $b->[1] <=> $a->[1] } @splitlist;
         for ( my $i=$start; ($i<($start+$stop) && ($i<$#sortlist)); $i++ )
         {
            push (@allrows, {
               urn    => $sortlist[$i]->[0],
               status => ''
            });
         }
         $hits = $#sortlist + 1;
      }
   }
   
   my $responseContainer =
         "<hits ".
         "xmlns=\"$self->{'odlsearch_xmlnsprefix'}/responsecontainer\" ".
         "xsi:schemaLocation=\"$self->{'odlsearch_xmlnsprefix'}/responsecontainer ".
         "$self->{'odlsearch_xmlnsprefix'}/responsecontainer.xsd\">".
         $hits.
         "</hits>";

   ( \@allrows, $resumptionToken, $responseContainer, $metadataPrefix );
}


# forward request to ListRecords handler
sub Archive_ListIdentifiers
{
   my ($self, $set, $from, $until, $resumptionToken) = @_;

   $self->Archive_ListRecords ($set, $from, $until, '', $resumptionToken);
}


# insert a new record into the archive
sub Archive_PutRecord
{
   my ($self, $identifier, $datestamp, $metadata, $metadataPrefix,
       $metadataNamespace, $metadataSchema, $about, $status, @sets) = @_;
       
   # parse input record
   my $parser = new Pure::EZXML;
   my $document = $parser->parse ($metadata)->getDocumentElement;
   
   my $identifiernodes = $document->getElementsByTagName ('identifier', 0);
   my $relationnodes = $document->getElementsByTagName ('relation', 0);
   if (($identifiernodes->getLength > 0) && ($relationnodes->getLength > 0))
   {
      # get ids
      my $subject = $identifiernodes->item(0)->getChildNodes->toString;
      my $object = $relationnodes->item(0)->getChildNodes->toString;

      # add new entry to table
      $self->{'dbh'}->do ("replace into $self->{'table'} set ".
                          "subject=\'$subject\', ".
                          "object=\'$object\', ".
                          "datestamp=\'$datestamp\'");
   }
   
   '0';
}


# main entry routine
sub Run
{
   my ($self) = @_;
   
   # create cgi object if necessary
   if (! exists $self->{'cgi'})
   {
      $self->{'cgi'} = new Pure::EZCGI;
   }
                  
   # check access control list
   my $ipaddress = $ENV{'REMOTE_ADDR'}; 
   if ((! defined $ipaddress) ||
       (($self->{'cgi'}->param ('verb') eq 'PutRecord') && (exists $self->{'allowwrite'}->{$ipaddress})) ||
       (exists $self->{'allowread'}->{$ipaddress}))
   { $self->SUPER::Run (); }
   else
   { print $self->Error ('forbidden operation', 403); }
}

1;

