//  ----------------------------------------------------------------------
// | Open Archives Initiative Repository Explorer - version 2.0b2-1.44    |
// | Hussein Suleman                                                      |
// | May 2002                                                             |
//  ----------------------------------------------------------------------
// |  Virginia Polytechnic Institute and State University                 |
// |  Department of Computer Science                                      |
// |  Digital Libraries Research Laboratory                               |
//  ----------------------------------------------------------------------

// ======================================================================
//  Module     : comply
//  Purpose    : automatic tests to check an archive for OAI compliance
// ======================================================================

#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "config.h"
#include "language.h"
#include "taglist.h"
#include "validate.h"
#include "xml.h"

// --------------------------------------------------------------------------

// global variables for storing parameters across tests
char repositoryName[2048] = "";
char protocolVersion[2048] = "";
char baseURL[2048] = "";
char adminEmail[2048] = "";
char metadataFormat[2048] = "";
char set[2048] = "";
char identifier[2048] = "";
char setSpec[10000] = "";
char set_resumptionToken[2048] = "";
char identifier_resumptionToken[2048] = "";
char record_resumptionToken[2048] = "";
char granularity[2048] = "";
int count = 1;

// error message pointer
char *errormsg;

// to update global variables or not
#define NO_UPDATE 0
#define UPDATE_VARIABLES 1

// to allow errors or not or maybe not to care at all
#define NO_ERROR 0
#define YES_ERROR 1
#define YES_OR_NO_ERROR 2

// ----------------------------------------------------------------------
//  Function   : DumpFile
// ----------------------------------------------------------------------
//  Purpose    : Outputs a raw XML file
//  Parameters : filename
//  Result     : None
// ----------------------------------------------------------------------

void DumpFile ( char *filename )
{
   FILE *rawf = fopen (filename, "r");
   char achar;
   cout << "------ Start of XML Response ------\n";
   while (!feof (rawf))
   {
      achar = getc (rawf);
      if (!feof (rawf)) 
      {
         if (achar == '<')
            cout << "&lt;";
         else if (achar == '>')
            cout << "&gt;";
         else if (achar == '&')
            cout << "&amp;";
         else
            cout << achar;
      }
   }
   fclose (rawf);
   cout << "------- End of XML Response -------\n";
}

// ----------------------------------------------------------------------
//  Function   : Process2
// ----------------------------------------------------------------------
//  Purpose    : Perform a test by issuing a request and checking the
//               results
//  Parameters : 1. base url of archive
//               2. verb being tested
//               3. parameters to be used for testing
//               4. do we allow errors ?
//               5. do we update global variables ?
//               6. what error to look for
//  Result     : Time/date string
// ----------------------------------------------------------------------

int Process2 ( char *archiveurl, char *verb, char *parameters, int error, int first, char *errorcode )
{
   char shellcommand[2048], oaiurl[2048], xsvcommand[2048];
   char datafilename[L_tmpnam];
   char errorfilename[L_tmpnam];
   char xsvfilename[L_tmpnam];

   tmpnam (datafilename);
   tmpnam (errorfilename);
   tmpnam (xsvfilename);
   
   if (first == UPDATE_VARIABLES)
   {
      if (strcmp (verb, "Identify") == 0)
      {
         repositoryName[0] = 0;
         protocolVersion[0] = 0;
         baseURL[0] = 0;
         adminEmail[0] = 0;
         strcpy (granularity, "YYYY-MM-DD");
      }
      else if (strcmp (verb, "ListMetadataFormats") == 0)
      {
         metadataFormat[0] = 0;
      }
      else if (strcmp (verb, "ListSets") == 0)
      {
         set[0] = 0;
         set_resumptionToken[0] = 0;
      }
      else if (strcmp (verb, "ListIdentifiers") == 0)
      {
         identifier[0] = 0;
         identifier_resumptionToken[0] = 0;
      }
      else if (strcmp (verb, "ListRecords") == 0)
      {
         record_resumptionToken[0] = 0;
      }
      else if (strcmp (verb, "GetRecord") == 0)
      {
         setSpec[0] = 0;
      }
   }

   sprintf (oaiurl, "%s?verb=%s%s", archiveurl, verb, parameters);

   cout << "URL : " << oaiurl << "\n"; 
   cout.flush ();

   sprintf (shellcommand, "%s -error_file=%s -source \"%s\" > %s", lynxpath, errorfilename, oaiurl, datafilename);

   system (shellcommand);

   char linebuffer[2048];
   int statuscode;
   fstream errorfile (errorfilename, ios::in);
   errorfile.getline (linebuffer, 2048);
   errorfile >> linebuffer >> statuscode;
   errorfile.getline (linebuffer, 2048);

   if (statuscode != 200)
   {
      static char str[2048];
      sprintf (str, "**** [%s] %s : %d / %s", Translate ("ERROR"), Translate ("Unexpected HTTP Error"), statuscode, linebuffer);
      errormsg = str;
      DumpFile (datafilename);
      unlink (xsvfilename);
      unlink (datafilename);
      unlink (errorfilename);
      return 0;
   }

   char *metadataPrefix;
   if (strstr (parameters, "metadataPrefix=oai_dc") != NULL)
      metadataPrefix = "oai_dc";
   else
      metadataPrefix = "";
   
   if (ValidateXerces ("2.0", datafilename, xsvfilename, verb, metadataPrefix, 1))
   {
      static char str[2048];
      sprintf (str, "**** [%s] %s", Translate ("ERROR"), Translate ("XML Schema validation failed"));
      errormsg = str;
      DumpFile (datafilename);
      unlink (xsvfilename);
      unlink (datafilename);
      unlink (errorfilename);
      return 0;
   }

   TagList tl, *p;
   if (!(runparser (&tl, datafilename)))
   {
      static char str[2048];
      sprintf (str, "**** [%s] %s", Translate ("ERROR"), Translate ("XML parsing errors"));
      errormsg = str;
      DumpFile (datafilename);
      unlink (xsvfilename);
      unlink (datafilename);
      unlink (errorfilename);
      return 0;
   }
   else
   {
      // check for error code
      p = tl.head->head;
      int ret = 0;
      while (p != NULL)
      {
         if (p->type == TAG)
         {
            if (strcmp (p->tag, "error") == 0)
            {
               if (((error == YES_ERROR) || (error == YES_OR_NO_ERROR)) &&
                   (strstr (errorcode, p->attr->Search ("code")) != NULL))
               {
                  unlink (xsvfilename);
                  unlink (datafilename);
                  unlink (errorfilename);  
                  ret = 1;
               }
               else if (((error == NO_ERROR) && (p->attr->Search ("code") != NULL)) ||
                        (((error == YES_ERROR) || (error == YES_OR_NO_ERROR)) &&
                         (strstr (errorcode, p->attr->Search ("code")) == NULL)))
               {
                  static char str[2048];
                  sprintf (str, "**** [%s] %s : %s", Translate ("ERROR"), Translate ("Error tag found but not expected"), p->attr->Search ("code"));
                  errormsg = str;
                  DumpFile (datafilename);
                  unlink (xsvfilename);
                  unlink (datafilename);
                  unlink (errorfilename);
                  return 0;
               }
            }
         }
         p = p->next;
      }
      if (ret == 1) { return 1; }
      
      if (error == YES_ERROR)
      {
         static char str[2048];
         sprintf (str, "**** [%s] %s : %s", Translate ("ERROR"), Translate ("Error tag expected but not found"), errorcode);
         errormsg = str;
         DumpFile (datafilename);
         unlink (xsvfilename);
         unlink (datafilename);
         unlink (errorfilename);
         return 0;
      }
      
      TagList *q = tl.head->Search (verb);
      if (q == NULL)
      {
         static char str[2048];
         sprintf (str, "**** [%s] %s : %s", Translate ("ERROR"), Translate ("Cannot find container corresponding to the verb"), verb);
         errormsg = str;
         DumpFile (datafilename);
         unlink (xsvfilename);
         unlink (datafilename);
         unlink (errorfilename);
         return 0;
      }
   
      if (first == UPDATE_VARIABLES)
      {
         if (strcmp (verb, "Identify") == 0)
         {
            p = q->Search ("repositoryName");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (repositoryName, p->head->tag);
            p = q->Search ("protocolVersion");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (protocolVersion, p->head->tag);
            p = q->Search ("baseURL");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (baseURL, p->head->tag);
            p = q->Search ("adminEmail");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (adminEmail, p->head->tag);
            p = q->Search ("granularity");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (granularity, p->head->tag);
         }
         else if (strcmp (verb, "ListMetadataFormats") == 0)
         {
            p = q->head;
            while (p)
            {
               if ((p->type == TAG) && (strcmp (p->tag, "metadataFormat") == 0))
               {
                  p = p->Search ("metadataPrefix");
                  if ((p) && (p->head) && (p->head->type == TEXT) && 
                      ((strlen (metadataFormat) == 0) || (strcmp (metadataFormat, "oai_dc") == 0)))
                     strcpy (metadataFormat, p->head->tag);
               }
               p = p->next;
            }
         
//            p = q->Search ("metadataFormat");
//            if (p)
//            {
//               p = p->Search ("metadataPrefix");
//               if ((p) && (p->head) && (p->head->type == TEXT))
//                  strcpy (metadataFormat, p->head->tag);
//            }            
         }
         else if (strcmp (verb, "ListSets") == 0)
         {
            p = q->Search ("set");
            if (p)
            {
               p = p->Search ("setSpec");
               if ((p) && (p->head) && (p->head->type == TEXT))
                  strcpy (set, p->head->tag);
            }
            p = q->Search ("resumptionToken");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (set_resumptionToken, p->head->tag);
         }
         else if (strcmp (verb, "ListIdentifiers") == 0)
         {
            p = q->Search ("header");
            if (p)
            {
               p = p->Search ("identifier");
               if ((p) && (p->head) && (p->head->type == TEXT))
               {
                  strcpy (identifier, p->head->tag);
               }
            }
            p = q->Search ("resumptionToken");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (identifier_resumptionToken, p->head->tag);
         }
         else if (strcmp (verb, "ListRecords") == 0)
         {
            p = q->Search ("resumptionToken");
            if ((p) && (p->head) && (p->head->type == TEXT))
               strcpy (record_resumptionToken, p->head->tag);
         }
         else if (strcmp (verb, "GetRecord") == 0)
         {
            p = q->Search ("record");
            if (p)
            {
               p = p->Search ("header");
               if (p)
               {
                  p = p->head;
                  while (p != NULL)
                  {
                     if ((p->type == TAG) && (strcmp (p->tag, "setSpec") == 0))
                     {
                        if ((p->head) && (p->head->type == TEXT))
                        {
                           strcat (setSpec, p->head->tag);
                           strcat (setSpec, ",");
                        }
                     }
                     p = p->next;
                  }
               }
            }
         }
      }
   }
   
   unlink (xsvfilename);
   unlink (datafilename);
   unlink (errorfilename);  
   
   return 1;
}

// ----------------------------------------------------------------------
//  Function   : Process
// ----------------------------------------------------------------------
//  Purpose    : perform test and output errors
//  Parameters : 1. description of verb being tested
//               2. base url of archive
//               3. verb being tested
//               4. parameters to be used for testing
//               5. do we allow errors ?
//               6. do we update global variables ?
//               7. error code string
//  Result     : Time/date string
// ----------------------------------------------------------------------

int Process ( char *descript, char *archiveurl, char *verb, char *parameters, int error, int first, char *errorcode )
{   
   cout << "\n(" << count++ << ") " << Translate ("Testing") << " : " << descript << "\n";
   
   errormsg = NULL;
   
   int status = Process2 (archiveurl, verb, parameters, error, first, errorcode);

   if (status)
   {
      cout << Translate ("Test Result") << " : " << Translate ("OK") << "\n";
      cout.flush ();
      return 0;
   }
   else
   {
      cout << Translate ("Test Result") << " : " << Translate ("FAIL!") << "\n";
      if (errormsg)
         cout << errormsg << "\n";
      cout.flush ();
      return 1;
   }
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

int main ( int argc, char **argv )
{
   char str[2048];
   int totalerrors = 0;
   int result;
   
   if (argc != 3)
   {
      cout << "Usage: comply <oai-url> <language>\n\n";
      exit (2);
   }
   
   SetLanguage (argv[2]);

   cout << "Open Archives Initiative :: " << Translate ("Protocol for Metadata Harvesting")
        << " v" << reprotocolversion << "\n" 
        << Translate ("RE Protocol Tester") << " " << reversion << " :: Virginia Tech DLRL :: "
        << Translate (redate) << "\n";

// IDENTIFY

   result = Process ("Identify", argv[1], "Identify", "", NO_ERROR, UPDATE_VARIABLES, "");
   totalerrors += result;
   if (result == 0)
   {
      if (strlen (repositoryName) > 0)
         cout << "---- [ " << Translate ("Repository Name") 
              << " = " << repositoryName << " ]\n";
      else
      {
         totalerrors++;
         cout << "**** [" << Translate ("ERROR") << "] " 
              << Translate ("Repository Name missing !") << "\n";
      }
      if (strlen (protocolVersion) > 0)
         cout << "---- [ " << Translate ("Protocol Version")
              << " = " << protocolVersion << " ]\n";
      else
      {
         totalerrors++;
         cout << "**** [" << Translate ("ERROR") << "] "
              << Translate ("Protocol Version missing !") << "\n";
      }
      if (strlen (baseURL) > 0)
         cout << "---- [ " << Translate ("Base URL") 
              << " = " << baseURL << " ]\n";
      else
      {
         totalerrors++;
         cout << "**** [" << Translate ("ERROR") << "] "
              << Translate ("Base URL missing !") << "\n";
      }
      if (strlen (adminEmail) > 0)
         cout << "---- [ " << Translate ("Admin Email") 
              << " = " << adminEmail << " ]\n";
      else
      {
         totalerrors++;
         cout << "**** [" << Translate ("ERROR") << "] "
              << Translate ("Admin Email missing !") << "\n";
      }
      if (strlen (granularity) > 0)
         cout << "---- [ " << Translate ("Granularity") 
              << " = " << granularity << " ]\n";
      else
      {
         strcpy (granularity, "YYYY-MM-DD");
      }
   }
   
   totalerrors += Process ("Identify (illegal_parameter)", argv[1], "Identify", "&test=test", YES_ERROR, NO_UPDATE, "badArgument");

// LIST METADATA FORMATS 1

   result = Process ("ListMetadataFormats", argv[1], "ListMetadataFormats", "", NO_ERROR, UPDATE_VARIABLES, "");
   totalerrors += result;
   if (result == 0)
   {
      if (strcmp (metadataFormat, "oai_dc") == 0)
         cout << "---- [ " << Translate ("Only oai_dc supported")
              << " ]\n";
      else if (strlen (metadataFormat) > 0)
         cout << "---- [ " << Translate ("Sample Metadata Format")
              << " = " << metadataFormat << " ]\n";
      else
      {
         totalerrors++;
         cout << "**** [" << Translate ("ERROR") << "] "
              << Translate ("Metadata format missing !") << "\n";
      }
   }

// LIST SETS

   result = Process ("ListSets", argv[1], "ListSets", "", YES_OR_NO_ERROR, UPDATE_VARIABLES, "noSetHierarchy,badArgument");
   totalerrors += result;
   if (result == 0)
   {
      if (strlen (set) > 0)
         cout << "---- [ " << Translate ("Sample Set Spec") 
              << " = " << set << " ]\n";

      if (strlen (set_resumptionToken) > 0)
      {
         cout << "---- [ " << Translate ("Set Resumption Token") 
              << " = " << set_resumptionToken << " ]\n";
         sprintf (str, "&resumptionToken=%s", set_resumptionToken);
         totalerrors += Process ("ListSets (resumptionToken)", argv[1], "ListSets", str, NO_ERROR, NO_UPDATE, "");
      }
   }

// LIST IDENTIFIERS

   result = Process ("ListIdentifiers (oai_dc)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc", NO_ERROR, UPDATE_VARIABLES, "");
   totalerrors += result;
   if (result == 0)
   {
      if (strlen (identifier) > 0)
         cout << "---- [ " << Translate ("Sample Identifier") 
              << " = " << identifier << " ]\n";
      else
      {
         totalerrors++;
         cout << "**** [" << Translate ("ERROR") << "] "
              << Translate ("Identifier missing !") << "\n";
      }

      if (strlen (identifier_resumptionToken) > 0)
      {
         cout << "---- [ " << Translate ("Identifier Resumption Token")
              << " = " << identifier_resumptionToken << " ]\n";
         sprintf (str, "&resumptionToken=%s", identifier_resumptionToken);
         totalerrors += Process ("ListIdentifiers (resumptionToken)", argv[1], "ListIdentifiers", str, NO_ERROR, NO_UPDATE, "");

         sprintf (str, "&resumptionToken=%s&metadataPrefix=oai_dc", identifier_resumptionToken);
         totalerrors += Process ("ListIdentifiers (resumptionToken, oai_dc)", argv[1], "ListIdentifiers", str, YES_ERROR, NO_UPDATE, "badArgument");
      }
   }

   totalerrors += Process ("ListIdentifiers (oai_dc, from/until)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2000-01-01&until=2000-01-01", YES_OR_NO_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");

   if (strlen (set) > 0)
   {
      sprintf (str, "&metadataPrefix=oai_dc&set=%s&from=2000-01-01&until=2000-01-01", set);
      totalerrors += Process ("ListIdentifiers (oai_dc, set, from/until)", argv[1], "ListIdentifiers", str, YES_OR_NO_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");
   }
 
   sprintf (str, "&metadataPrefix=oai_dc&set=really_wrong_set&from=some_random_date&until=some_random_date");
   totalerrors += Process ("ListIdentifiers (oai_dc, illegal_set, illegal_from/until)", argv[1], "ListIdentifiers", str, YES_ERROR, NO_UPDATE, "badArgument,noSetHierarchy");

   totalerrors += Process ("ListIdentifiers (oai_dc, from granularity != until granularity)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2001-01-01&until=2002-01-01T00:00:00Z", YES_ERROR, NO_UPDATE, "badArgument");

   totalerrors += Process ("ListIdentifiers (oai_dc, from > until)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2000-01-01&until=1999-01-01", YES_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");

   totalerrors += Process ("ListIdentifiers ()", argv[1], "ListIdentifiers", "", YES_ERROR, NO_UPDATE, "badArgument");

   if ((strlen (metadataFormat) > 0) && (strcmp (metadataFormat, "oai_dc") != 0))
   {
      sprintf (str, "&metadataPrefix=%s", metadataFormat);
      totalerrors += Process ("ListIdentifiers (metadataPrefix)", argv[1], "ListIdentifiers", str, YES_OR_NO_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");
   }

   totalerrors += Process ("ListIdentifiers (illegal_mdp)", argv[1], "ListIdentifiers", "&metadataPrefix=illegal_mdp", YES_ERROR, NO_UPDATE, "cannotDisseminateFormat,badArgument");

   totalerrors += Process ("ListIdentifiers (mdp, mdp)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&metadataPrefix=oai_dc", YES_ERROR, NO_UPDATE, "badArgument");

   
// GRANULARITY TESTS

   totalerrors += Process ("ListIdentifiers (oai_dc, from YYYY-MM-DD)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2001-01-01", YES_OR_NO_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");
   
   if (strcmp (granularity, "YYYY-MM-DDThh:mm:ssZ") == 0)
   {
      totalerrors += Process ("ListIdentifiers (oai_dc, from YYYY-MM-DDThh:mm:ssZ)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2001-01-01T00:00:00Z", YES_OR_NO_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");
   }
   else
   {
      totalerrors += Process ("ListIdentifiers (oai_dc, from YYYY-MM-DDThh:mm:ssZ)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2001-01-01T00:00:00Z", YES_ERROR, NO_UPDATE, "badArgument");
   }
   
   totalerrors += Process ("ListIdentifiers (oai_dc, from YYYY)", argv[1], "ListIdentifiers", "&metadataPrefix=oai_dc&from=2001", YES_ERROR, NO_UPDATE, "badArgument");
   

// LIST METADATA FORMATS 2

   if (strlen (identifier) > 0)
   {
      sprintf (str, "&identifier=%s", identifier);
      result = Process ("ListMetadataFormats (identifier)", argv[1], "ListMetadataFormats", str, NO_ERROR, UPDATE_VARIABLES, "");
      totalerrors += result;

      if (result == 0)
      {
         if (strcmp (metadataFormat, "oai_dc") == 0)
            cout << "---- [ " << Translate ("Only oai_dc supported")
                 << " ]\n";
         else if (strlen (metadataFormat) > 0)
            cout << "---- [ " << Translate ("Sample Metadata Format")
                 << " = " << metadataFormat << " ]\n";
         else
         {
            totalerrors++;
            cout << "**** [" << Translate ("ERROR") << "] "
                 << Translate ("Metadata format missing !") << "\n";
         }
      }
   }

   sprintf (str, "&identifier=really_wrong_id");
   totalerrors += Process ("ListMetadataFormats (illegal_id)", argv[1], "ListMetadataFormats", str, YES_ERROR, NO_UPDATE, "idDoesNotExist,badArgument");

// GET RECORD

   if (strlen (identifier) > 0)
   {
      if ((strlen (metadataFormat) > 0) && (strcmp (metadataFormat, "oai_dc") != 0))
      {
         sprintf (str, "&identifier=%s&metadataPrefix=%s", identifier, metadataFormat);
         totalerrors += Process ("GetRecord (identifier, metadataPrefix)", argv[1], "GetRecord", str, NO_ERROR, NO_UPDATE, "");
      }

      sprintf (str, "&identifier=%s&metadataPrefix=oai_dc", identifier);
      totalerrors += Process ("GetRecord (identifier, oai_dc)", argv[1], "GetRecord", str, NO_ERROR, NO_UPDATE, "");

      sprintf (str, "&identifier=%s", identifier);
      totalerrors += Process ("GetRecord (identifier)", argv[1], "GetRecord", str, YES_ERROR, NO_UPDATE, "badArgument");

      sprintf (str, "&identifier=%s&metadataPrefix=really_wrong_mdp", identifier);
      totalerrors += Process ("GetRecord (identifier, illegal_mdp)", argv[1], "GetRecord", str, YES_ERROR, NO_UPDATE, "cannotDisseminateFormat,badArgument");
   }

   totalerrors += Process ("GetRecord (oai_dc)", argv[1], "GetRecord", "&metadataPrefix=oai_dc", YES_ERROR, NO_UPDATE, "badArgument");

   totalerrors += Process ("GetRecord (illegal_id, oai_dc)", argv[1], "GetRecord", "&identifier=really_wrong_id&metadataPrefix=oai_dc", YES_ERROR, NO_UPDATE, "idDoesNotExist,badArgument");

// LIST RECORDS

   result = Process ("ListRecords (oai_dc, from/until)", argv[1], "ListRecords", "&metadataPrefix=oai_dc&from=2000-01-01&until=2000-01-01", YES_OR_NO_ERROR, UPDATE_VARIABLES, "noRecordsMatch,badArgument");
   totalerrors += result;

   if (result == 0)
   {
      if (strlen (record_resumptionToken) > 0)
      {
         cout << "---- [ " << Translate ("Record Resumption Token") 
              << " = " << record_resumptionToken << " ]\n";
         sprintf (str, "&resumptionToken=%s", record_resumptionToken);
         totalerrors += Process ("ListRecords (resumptionToken)", argv[1], "ListRecords", str, NO_ERROR, NO_UPDATE, "");
      }
   }
   
   if ((strlen (metadataFormat) > 0) && (strcmp (metadataFormat, "oai_dc") != 0))
   {
      sprintf (str, "&metadataPrefix=%s&from=2000-01-01&until=2000-01-01", metadataFormat);
      totalerrors += Process ("ListRecords (metadataPrefix, from/until)", argv[1], "ListRecords", str, YES_OR_NO_ERROR, NO_UPDATE, "noRecordsMatch,badArgument");
   }

   sprintf (str, "&metadataPrefix=oai_dc&set=really_wrong_set&from=some_random_date&until=some_random_date");
   totalerrors += Process ("ListRecords (oai_dc, illegal_set, illegal_from/until)", argv[1], "ListRecords", str, YES_ERROR, NO_UPDATE, "badArgument,noSetHierarchy");

   totalerrors += Process ("ListRecords", argv[1], "ListRecords", "", YES_ERROR, NO_UPDATE, "badArgument");

   totalerrors += Process ("ListRecords (oai_dc, from granularity != until granularity)", argv[1], "ListRecords", "&metadataPrefix=oai_dc&from=2001-01-01&until=2002-01-01T00:00:00Z", YES_ERROR, NO_UPDATE, "badArgument");

// SET CORRESPONDENCE

   if (strlen (set) > 0)
   {
      sprintf (str, "&metadataPrefix=oai_dc&set=%s", set);
      result = Process ("ListIdentifiers (oai_dc, set)", argv[1], "ListIdentifiers", str, YES_OR_NO_ERROR, UPDATE_VARIABLES, "noRecordsMatch,badArgument");
      totalerrors += result;
 
      if (result == 0)
      {
         if (strlen (identifier) > 0)
         {
            sprintf (str, "&identifier=%s&metadataPrefix=oai_dc", identifier);
            result = Process ("GetRecord (identifier, oai_dc)", argv[1], "GetRecord", str, NO_ERROR, UPDATE_VARIABLES, "");
            totalerrors += result;

            if (result == 0)
            {
               if (strstr (setSpec, set) != NULL)
                  cout << "---- [ " << Translate ("Found setSpec in header") 
                       << " ]\n";
               else
               {
                  totalerrors++;
                  cout << "**** [" << Translate ("ERROR") << "] "
                       << Translate ("Could not find setSpec in header !")
                       << " set = " << set << ", setSpecs=" << setSpec << "\n";
               }
            }
         }
      }
   }

// ILLEGAL VERB

   totalerrors += Process ("IllegalVerb", argv[1], "IllegalVerb", "", YES_ERROR, NO_UPDATE, "badVerb");

   cout << "\n\n---- " << Translate ("Total Errors") << " : "
        << totalerrors << "\n";
         
   if (totalerrors > 0)
      exit (1);
   else
      exit (0);
}

	