//  ----------------------------------------------------------------------
// | Open Archives Initiative Repository Explorer - version 1.0-1.1       |
// | Hussein Suleman                                                      |
// | January 2001                                                         |
//  ----------------------------------------------------------------------
// |  Virginia Polytechnic Institute and State University                 |
// |  Department of Computer Science                                      |
// |  Digital Libraries Research Laboratory                               |
//  ----------------------------------------------------------------------

// ======================================================================
//  Module     : parser
//  Purpose    : Parse valid XML
// ======================================================================

#include <fstream.h>
#include <iostream.h>
#include <string.h>

#include "escape.h"
#include "language.h"
#include "parserodl.h"

// forward declaration
void ProcessXResponseContainer ( TagList *tl );

// ----------------------------------------------------------------------
//  Function   : ErrorX{IllegalTag,..}
// ----------------------------------------------------------------------
//  Purpose    : Output error messages
//  Parameters : (variable)
//  Result     : (none)
// ----------------------------------------------------------------------

void ErrorXIllegalTag ( char *tag, char *parent, char *correct )
{
   cout << Translate ("Error") << ": " 
        << Translate ("Illegal tag")
        << " : ";
   if (parent)
      cout << "&lt;" << parent << "&gt;/";
   cout << "&lt;" << tag << "&gt;";
   if (correct)
      cout << " <> {" << correct << "}";
   cout << "<p>";
}

void ErrorXMissingField ( char *tag, char *parent )
{
   cout << Translate ("Error") << ": "
        << Translate ("Missing field")
        << " : &lt;" << parent << "&gt;/&lt;" << tag << "&gt;"        
        << "<p>";
}

void ErrorXTextWhereTag ( char *text )
{
   cout << Translate ("Error") << ": " 
        << Translate ("Text found where tag expected")
        << " : ";
   Escape (text);
   cout << "<p>";
}

void ErrorXTagNotFound ( char *tag )
{
   cout << Translate ("Error") << ": " 
        << Translate ("Tag expected but not found")
        << " : &lt;" << tag << "&gt;<p>";
}

void ErrorXOnlyOneInstance ( char *tag )
{
   cout << Translate ("Error") << ": "
        << Translate ("Only one instance allowed")
        << " : &lt;" << tag << "&gt;<p>";
}

void ErrorXTextNotFound ( char *tag )
{
   cout << Translate ("Error") << ": "
        << Translate ("Text expected for field but not found")
        << " : &lt;" << tag << "&gt;<p>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXIRecord
// ----------------------------------------------------------------------
//  Purpose    : Process identity tags
//  Parameters : 1. Tag pointer
//               2. Depth in nested structure
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXIRecord ( TagList *tl, int depth=0 )
{
   // indent
   for ( int i=0; i<depth*3; i++ )
      cout << " ";
      
   // output tag
   Escape (tl->tag);
   cout << ": ";
   tl->attr->List ();
   
   // empty tag
   if (tl->head == NULL)
   {
      cout << "\n";
   }
   
   // tag containing text
   else if (tl->head->type == TEXT)
   {
      Escape (tl->head->tag);
      cout << "\n";
   }
   
   // tag with child elements
   else
   {
      cout << "\n";
      TagList *p = tl->head;
      while (p != NULL)
      {
         ProcessXIRecord (p, depth+1);
         p = p->next;
      }
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXI
// ----------------------------------------------------------------------
//  Purpose    : Process Identify tag
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXI ( TagList *tl, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("Archive Self-Description") 
        << "</h2><p><pre>";
   
   char *repositoryName = NULL;
   char *baseURL = NULL;
   char *protocolVersion = NULL;
   char *adminEmail = NULL;
   
   // check for mandatory fields
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "repositoryName") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               repositoryName = p->head->tag;
         }
         else if (strcmp (p->tag, "baseURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               baseURL = p->head->tag;
         }
         else if (strcmp (p->tag, "protocolVersion") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               protocolVersion = p->head->tag;
         }
         else if (strcmp (p->tag, "adminEmail") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               adminEmail = p->head->tag;
         }
      }
      p = p->next;
   }
   
   // output metadata
   cout << "<table border=1 cellspacing=2 cellpadding=5>";
   if (repositoryName)
      cout << "<tr><th align=left>" << Translate ("Repository Name") 
           << "</th><td>" << repositoryName << "</td></tr>";
   else
   {
      cout << "<tr><td colspan=2>";
      ErrorXMissingField ("repositoryName", "Identify");
      cout << "</td></tr>";
   }
   if (baseURL)
      cout << "<tr><th align=left>" << Translate ("Base URL") 
           << "</th><td>" << baseURL << "</td></tr>";
   else
   {
      cout << "<tr><td colspan=2>";
      ErrorXMissingField ("baseURL", "Identify");
      cout << "</td></tr>";
   }
   if (protocolVersion)
      cout << "<tr><th align=left>" << Translate ("Protocol Version")
           << "</th><td>" << protocolVersion << "</td></tr>";
   else
   {
      cout << "<tr><td colspan=2>";
      ErrorXMissingField ("protocolVersion", "Identify");
      cout << "</td></tr>";
   }
   if (adminEmail)
      cout << "<tr><th align=left>" << Translate ("Admin Email")
           << "</th><td>" << adminEmail << "</td></tr>";
   else
   {
      cout << "<tr><td colspan=2>";
      ErrorXMissingField ("adminEmail", "Identify");
      cout << "</td></tr>";
   }
   cout << "<tr><th align=left>" << Translate ("Other Information")
        << "</th><td><pre>";

   // process optional descriptions
   p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "description") == 0)
            ProcessXIRecord (p);
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            if ((strcmp (p->tag, "responseDate") != 0) &&
                (strcmp (p->tag, "requestURL") != 0) &&
                (strcmp (p->tag, "repositoryName") != 0) &&
                (strcmp (p->tag, "baseURL") != 0) &&
                (strcmp (p->tag, "protocolVersion") != 0) &&
                (strcmp (p->tag, "adminEmail") != 0))
               ErrorXIllegalTag (p->tag, "Identify", "repositoryName,baseURL,protocolVersion,adminEmail,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }

   cout << "</pre></td></tr>";
   cout << "</table>";

   cout << "</pre>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXIdentify
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response for Identify
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXIdentify ( TagList *tl, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("Identify");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "Identify") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "Identify");
   else
      ProcessXI (tl->head, requestURL, responseDate);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLSResumptionToken
// ----------------------------------------------------------------------
//  Purpose    : Process resumptionToken tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLSResumptionToken ( TagList *tl )
{
   TagList *p = tl->head;
   while (p != NULL)
   {
      if ((p->type == TEXT) && (strlen (p->tag) > 0))
      {
         cout << "\n <a href=\"\" onClick=\"FillInRT (\'ListSets\', \'" << p->tag << "\');"
              << "return false\">" 
              << Translate ("Resume from") << " [" << p->tag << "] </a> ";
         cout << "\n\n";
      }
      p = p->next;
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLSSet
// ----------------------------------------------------------------------
//  Purpose    : Process set tag
//  Parameters : 1. Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLSSet ( TagList *tl )
{
   char *setSpec = NULL;
   char *setName = NULL;

   // find fields in metadata
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "setSpec") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
            {
               if (setSpec)
                  ErrorXOnlyOneInstance ("setSpec");
               else
                  setSpec = p->head->tag;
            }
            else
               ErrorXTextNotFound ("setSpec");
         }
         else if (strcmp (p->tag, "setName") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
            {
               if (setName)
                  ErrorXOnlyOneInstance ("setName");
               else
                  setName = p->head->tag;
            }
            else
               ErrorXTextNotFound ("setName");
         }
         else
            ErrorXIllegalTag (p->tag, "set", "setSpec,setName");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }

   // output set link and description
   if (setSpec == NULL)
      ErrorXMissingField ("setSpec", "set");
   else if (setName == NULL)
      ErrorXMissingField ("setSpec", "set");
   else
   {   
      cout << "<a href=\"\" onClick=\"FillInLI (\'" << setSpec << "\');"
           << "return false\">";
      Escape (setName);
      cout << "</a>"; 
      cout << "<p>";
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLS
// ----------------------------------------------------------------------
//  Purpose    : Process ListSets tag
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLS ( TagList *tl, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("List of Sets") << "</h2><i>"
        << Translate ("Click on the link to list the contents")
        << "</i><p><pre>";
   
   // iterate over all sets
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "set") == 0)
            ProcessXLSSet (p);
         else if (strcmp (p->tag, "resumptionToken") == 0)
            ProcessXLSResumptionToken (p);
         else if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            ErrorXIllegalTag (p->tag, "ListSets", "set,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }
   cout << "</pre>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXListSets
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response for List-Sets
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXListSets ( TagList *tl, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("ListSets");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "ListSets") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "ListSets");
   else
      ProcessXLS (tl->head, requestURL, responseDate);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLMFMetadataFormat
// ----------------------------------------------------------------------
//  Purpose    : Process metadataformat tag
//  Parameters : 1. Tag pointer
//               2. Pointer to full id
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLMFMetadataFormat ( TagList *tl, char *identifier )
{
   char *metadataPrefix = NULL;
   char *schema = NULL;
   char *metadataNamespace = NULL;

   // find fields in tag list
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "metadataPrefix") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
            {
               if (metadataPrefix)
                  ErrorXOnlyOneInstance ("metadataprefix");
               else
                  metadataPrefix = p->head->tag;
            }
            else
               ErrorXTextNotFound ("metadataPrefix");
         }
         else if (strcmp (p->tag, "schema") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
            {
               if (schema)
                  ErrorXOnlyOneInstance ("schema");
               else
                  schema = p->head->tag;
            }
            else
               ErrorXTextNotFound ("schema");
         }
         else if (strcmp (p->tag, "metadataNamespace") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
            {
               if (metadataNamespace)
                  ErrorXOnlyOneInstance ("metadataNamespace");
               else
                  metadataNamespace = p->head->tag;
            }
            else
               ErrorXTextNotFound ("metadataNamespace");
         }
         else
            ErrorXIllegalTag (p->tag, "metadataFormat", "metadataPrefix,schema,metadataNamespace");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }

   // output information about metadata format and links
   if (metadataPrefix == NULL)
      ErrorXMissingField ("metadataPrefix", "metadataFormat");
   else if (schema == NULL)
      ErrorXMissingField ("schema", "metadataFormat");
   else
   {   
      cout << Translate ("Prefix") << "=[" << metadataPrefix << "]<br>";
      if (metadataNamespace)
         cout << Translate ("NameSpace") << "=[" << metadataNamespace << "]<br>";
      cout << Translate ("Schema") << "=[" << "<a href=\"" << schema << "\">" << schema << "</a>]<p>";
   
      int metamatch = 0;
      fstream sfcmeta ("sfcmeta", ios::in);
      char ameta[1024];
      while (!(sfcmeta.eof ()))
      {
         sfcmeta.getline (ameta, sizeof (ameta));
         if (strcmp (ameta, metadataPrefix) == 0)
            metamatch = 1;
      }
   
      if (metamatch == 0)
      {
         cout << "[" << Translate ("Not a standard OAI metadata name")
              << "]";
      }
      
      if ((identifier) && (strlen (identifier) > 0))
      {
         cout << " <a href=\"\" onClick=\"FillInGR (\'" << metadataPrefix << "\',\'" << identifier << "\');"
              << "return false\">" << "[" 
              << Translate ("display record") << "]" << "</a> ";
      }

      cout << "<p>";
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLMF
// ----------------------------------------------------------------------
//  Purpose    : Process ListMetadataFormats tag
//  Parameters : 1. Tag pointer
//               2. Pointer to full id
//               3. requestURL pointer
//               4. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLMF ( TagList *tl, char *identifier, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("List of Metadata Formats") 
        << "</h2><i>" << Translate ("Click on the link to view schema")
        << "</i><p>";

   // iterate over metadata formats
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "metadataFormat") == 0)
            ProcessXLMFMetadataFormat (p, identifier);
         else if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            ErrorXIllegalTag (p->tag, "ListMetadataFormats", "metadataFormat,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXListMetadataFormats
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response for ListMetadataFormats
//  Parameters : 1. Tag pointer
//               2. Pointer to full id
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer

//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXListMetadataFormats ( TagList *tl, char *identifier, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("ListMetadataFormats");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "ListMetadataFormats") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "ListMetadataFormats");
   else
      ProcessXLMF (tl->head, identifier, requestURL, responseDate);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLIIdentifier
// ----------------------------------------------------------------------
//  Purpose    : Process identifier within identifiers tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLIIdentifier ( TagList *tl )
{
   TagList *p = tl->head;
   while (p != NULL)
   {
      if ((p->type == TEXT) && (strlen (p->tag) > 0))
      {
         cout << p->tag << " ";
         cout << "<a href=\"\" onClick=\"FillInGR (\'oai_dc\',\'" << p->tag << "\');"
              << "return false\">" << "[" 
              << Translate ("display record in Dublin Core") << "]" << "</a> ";
         cout << "<a href=\"\" onClick=\"FillInLMF (\'" << p->tag << "\');"
              << "return false\">" << "["
              << Translate ("display metadata formats") << "]" << "</a> ";
         cout << "<p>";
      }
      p = p->next;
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLIResumptionToken
// ----------------------------------------------------------------------
//  Purpose    : Process resumptionToken tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLIResumptionToken ( TagList *tl )
{
   TagList *p = tl->head;
   while (p != NULL)
   {
      if ((p->type == TEXT) && (strlen (p->tag) > 0))
      {
         cout << "\n <a href=\"\" onClick=\"FillInRT (\'ListIdentifiers\',\'" << p->tag << "\');"
              << "return false\">" 
              << Translate ("Resume from") << " [" << p->tag << "] </a> ";
         cout << "\n\n";
      }
      p = p->next;
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLI
// ----------------------------------------------------------------------
//  Purpose    : Process ListIdentifiers tag
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLI ( TagList *tl, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("List of Record Identifiers")
        << "</h2><i>" << Translate ("Select a link to view more information")
        << "</i><p><pre>";
        
   // iterate over identifiers
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "identifier") == 0)
            ProcessXLIIdentifier (p);
         else if (strcmp (p->tag, "resumptionToken") == 0)
            ProcessXLIResumptionToken (p);
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            ErrorXIllegalTag (p->tag, "ListIdentifiers", "identifier,resumptionToken,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }
   cout << "</pre>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXListIdentifiers
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response to ListIdentifiers
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXListIdentifiers ( TagList *tl, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("ListIdentifiers");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "ListIdentifiers") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "ListIdentifiers");
   else
      ProcessXLI (tl->head, requestURL, responseDate);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGRField
// ----------------------------------------------------------------------
//  Purpose    : Process fields of record
//  Parameters : 1. Tag pointer
//               2. Depth in nested structure
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGRField ( TagList *tl, int depth )
{
   // indent
   for ( int i=0; i<depth*3; i++ )
      cout << " ";
      
   // output tag and attributes   
   Escape (tl->tag); 
   cout << ": ";
   tl->attr->List (); 

   // empty tag
   if (tl->head == NULL)
      cout << "\n";
   
   // tag containing text
   else if (tl->head->type == TEXT)
   {
      // check for links within metadata
      int http = 0;
      if (strlen (tl->head->tag) > 7)
      {
         char protocol[8];
         for ( int j=0; j<7; j++ )
            protocol[j] = tl->head->tag[j];
         protocol[7] = 0;
         if (strcmp (protocol, "http://") == 0)
            http = 1;
      }
      if (http)
         cout << "<a href=\"" << tl->head->tag << "\">" << tl->head->tag << "</a>\n";
      else
      {
         Escape (tl->head->tag);
         cout << "\n";
      }
   }
   
   // tag with elements
   else
   {
      cout << "\n";
      TagList *p = tl->head;
      while (p != NULL)
      {
         ProcessXGRField (p, depth+1);
         p = p->next;
      }
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGRMetadata
// ----------------------------------------------------------------------
//  Purpose    : Process record metadata tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGRMetadata ( TagList *tl )
{
   cout << "<b>metadata:</b>\n";
   ProcessXGRField (tl, 1);
   cout << "\n";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGRAbout
// ----------------------------------------------------------------------
//  Purpose    : Process record about tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGRAbout ( TagList *tl )
{
   cout << "<b>about:</b>\n";
   ProcessXGRField (tl, 1);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGRHeader
// ----------------------------------------------------------------------
//  Purpose    : Process record header tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGRHeader ( TagList *tl )
{
   TagList *p = tl;
   TagList *identifier = NULL;
   TagList *datestamp = NULL;

   // iterate over fields
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "identifier") == 0)
         {
            if (identifier)
               ErrorXOnlyOneInstance ("identifier");
            else
               identifier = p;
         }
         else if (strcmp (p->tag, "datestamp") == 0)
         {
            if (datestamp)
               ErrorXOnlyOneInstance ("datestamp");
            else
               datestamp = p;
         }
         else if (strcmp (p->tag, "setSpec") != 0)
            ErrorXIllegalTag (p->tag, "header", "identifier,datestamp");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }

   // output header information
   cout << "\n\n<b>header:</b>\n";
   if (identifier)
   {
      if ((identifier->head) && (identifier->head->type == TEXT))
         cout << "  " << "identifier" << " : " 
              << identifier->head->tag << "\n";
      else
         ErrorXTextNotFound ("identifier");
   }
   else
      ErrorXMissingField ("identifier", "header");
   if (datestamp)
   {
      if ((datestamp->head) && (datestamp->head->type == TEXT))
         cout << "  " << "datestamp" 
              << " : " << datestamp->head->tag << "\n";
      else
         ErrorXTextNotFound ("datestamp");
   }
   else
      ErrorXMissingField ("datestamp", "header");
         
   // iterate over setSpec fields
   p = tl;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "setSpec") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               cout << "  " << "setSpec" << " : "
                    << p->head->tag << "\n";
         }
      }
      p = p->next;
   }
      
   cout << "\n";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGRRecord
// ----------------------------------------------------------------------
//  Purpose    : Process record tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGRRecord ( TagList *tl )
{
   TagList *p = tl->head;
   TagList *header = NULL;
   TagList *metadata = NULL;
   TagList *about = NULL;

   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "header") == 0)
         {
            if (header)
               ErrorXOnlyOneInstance ("header");
            else
               header = p->head;
         }
         else if (strcmp (p->tag, "metadata") == 0)
         {
            if (metadata)
               ErrorXOnlyOneInstance ("metadata");
            else
               metadata = p->head;
         }
         else if (strcmp (p->tag, "about") == 0)
         {
            if (about)
               ErrorXOnlyOneInstance ("about");
            else
               about = p->head;
         }
         else
            ErrorXIllegalTag (p->tag, "record", "header,metadata,about");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }

   if (header)
      ProcessXGRHeader (header);
   else
      ErrorXMissingField ("header", "record");
   if (metadata)
      ProcessXGRMetadata (metadata);
   if (about)
      ProcessXGRAbout (about);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGR
// ----------------------------------------------------------------------
//  Purpose    : Process GetRecord tag
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGR ( TagList *tl, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("List of Fields") << "</h2><pre>";
   
   // iterate over "records"
   TagList *p = tl->head;
   TagList *record = NULL;
   while (p != NULL)
   {  
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "record") == 0)
         {
            if (record)
               ErrorXOnlyOneInstance ("record");
            else
               record = p;
         }
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            ErrorXIllegalTag (p->tag, "GetRecord", "record,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }
   if (record)
      ProcessXGRRecord (record);
   else
      ErrorXMissingField ("record", "GetRecord");
   cout << "</pre>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXGetRecord
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response to GetRecord
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXGetRecord ( TagList *tl, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("GetRecord");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "GetRecord") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "GetRecord");
   else
      ProcessXGR (tl->head, requestURL, responseDate);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLRResumptionToken
// ----------------------------------------------------------------------
//  Purpose    : Process resumptionToken tag
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLRResumptionToken ( TagList *tl )
{
   TagList *p = tl->head;
   while (p != NULL)
   {
      if ((p->type == TEXT) && (strlen (p->tag) > 0))
      {
         cout << "\n <a href=\"\" onClick=\"FillInRT (\'ListRecords\',\'" << p->tag << "\');"
              << "return false\">" 
              << Translate ("Resume from") << " [" << p->tag << "] </a> ";
         cout << "\n\n";
      }
      p = p->next;
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXResponseContainer
// ----------------------------------------------------------------------
//  Purpose    : Process responsecontainer tags
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXResponseContainer ( TagList *tl )
{
   cout << "<hr><b>responsecontainer:</b>\n";
   ProcessXGRField (tl->head, 1);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXLR
// ----------------------------------------------------------------------
//  Purpose    : Process ListRecords tag
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXLR ( TagList *tl, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("List of Records") 
        << "</h2><i>" << Translate ("Select a link to view more information")
        << "</i><p><pre>";

   TagList *p = tl->head;
   while (p != NULL)
   {  
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "resumptionToken") == 0)
            ProcessXLRResumptionToken (p);
         else if (strcmp (p->tag, "record") == 0)
            ProcessXGRRecord (p);
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            ErrorXIllegalTag (p->tag, "ListRecords", "record,resumptionToken,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }
   cout << "</pre>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXListRecords
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response to ListRecords
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXListRecords ( TagList *tl, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("ListRecord");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "ListRecords") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "ListRecords");
   else
      ProcessXLR (tl->head, requestURL, responseDate);
}

// ----------------------------------------------------------------------
//  Function   : ProcessXPRIdentifier
// ----------------------------------------------------------------------
//  Purpose    : Process identifier within PutRecord response
//  Parameters : Tag pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXPRIdentifier ( TagList *tl )
{
   TagList *p = tl->head;
   while (p != NULL)
   {
      if ((p->type == TEXT) && (strlen (p->tag) > 0))
      {
         cout << p->tag << " ";
         cout << "<a href=\"\" onClick=\"FillInGR (\'oai_dc\',\'" << p->tag << "\');"
              << "return false\">" << "[" 
              << Translate ("display record in Dublin Core") << "]" << "</a> ";
         cout << "<a href=\"\" onClick=\"FillInLMF (\'" << p->tag << "\');"
              << "return false\">" << "["
              << Translate ("display metadata formats") << "]" << "</a> ";
         cout << "<p>";
      }
      p = p->next;
   }
}

// ----------------------------------------------------------------------
//  Function   : ProcessXPR
// ----------------------------------------------------------------------
//  Purpose    : Process PutRecord tag
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXPR ( TagList *tl, char **requestURL, char **responseDate )
{
   cout << "<h2>" << Translate ("Identifier of submitted record")
        << "</h2>"
        << "<p><pre>";
        
   // iterate over identifiers
   TagList *p = tl->head;
   while (p != NULL)
   {
      if (p->type == TAG)
      {
         if (strcmp (p->tag, "requestURL") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *requestURL = p->head->tag;
         }
         else if (strcmp (p->tag, "responseDate") == 0)
         {
            if ((p->head) && (p->head->type == TEXT))
               *responseDate = p->head->tag;
         }
         else if (strcmp (p->tag, "identifier") == 0)
            ProcessXPRIdentifier (p);
         else if (strcmp (p->tag, "responseContainer") == 0)
            ProcessXResponseContainer (p);
         else
            ErrorXIllegalTag (p->tag, "PutRecord", "identifier,responseContainer");
      }
      else
         ErrorXTextWhereTag (p->tag);
      p = p->next;
   }
   cout << "</pre>";
}

// ----------------------------------------------------------------------
//  Function   : ProcessXPutRecord
// ----------------------------------------------------------------------
//  Purpose    : Process OAI response to PutRecord
//  Parameters : 1. Tag pointer
//               2. requestURL pointer
//               3. responseDate pointer
//  Result     : (none)
// ----------------------------------------------------------------------

void ProcessXPutRecord ( TagList *tl, char **requestURL, char **responseDate )
{
   if (tl->head == NULL)
      ErrorXTagNotFound ("PutRecord");
   else if (tl->head->type == TEXT)
      ErrorXTextWhereTag (tl->head->tag);
   else if (strcmp (tl->head->tag, "PutRecord") != 0)
      ErrorXIllegalTag (tl->head->tag, NULL, "PutRecord");
   else
      ProcessXPR (tl->head, requestURL, responseDate);
}
